/* FCHT.h
 * Copyright (C) 2010, Rodrigo Canovas & Miguel A. Martinez-Prieto
 * all rights reserved.
 *
 * FrontCoding Compression on Hu-Tucker Encoding
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *
 * Contacting the author:
 *   Rodrigo Canovas:  yigorc@gmail.com
 *   Miguel A Martinez-Prieto:  migumar2@infor.uva.es
 */


#ifndef _FCHT_H__F
#define _FCHT_H__F

#include <iostream>
#include <fstream>
#include <string.h>

#include "libcdsutils/Array_F.h"
#include "VByte_F.h"
#include "hutucker.h"

using namespace std;

typedef struct
{
	int children[2];
	int symbol;
} Node;

class FCHT 
{	
	public:
		FCHT();
		FCHT(char *filename, size_t blocksize);
		FCHT(uchar ** words, uint n, size_t blocksize);

		/** Return the ID identifying the word w[0..len-1] */
		size_t locate(uchar *w, size_t len);
		/** Return the word identified by id */
		uchar * extract(size_t id);

		/** Returns the size of the structure in bytes */
		size_t getSize();
		/** Returns the number of words in the dictionary */
		size_t getLength();

		/** Stores the structure given a file pointer */
		void save(ofstream & fp);
		/** Load a FCHT data structure */
		static FCHT * load(ifstream & fp);
			
		~FCHT();

	protected:
		size_t length;	 // Number of elements in the dictionary
		size_t max_len;  // Max length URI

		Array_F *blocks;
		size_t blocksize;
		size_t nblocks;

		uchar *text;
		size_t uchars;
		BitString_F *tree;
		Node *HTtree;
		Tcode* HTcode;

		uchar *qstring;
		uchar *delta;
		uint leafsHT;

		/** Extracts (char-to-char) the URI starting at 'pos'
		    and compares it with respect to given URI.
		    Returns: 0  if equals.
			     >0 if 'uri' is before 'pos'.
			     <0 if 'uri' is after 'pos'.
		**/
		int extractAndCompare(uint pos, uchar* uri);
		uint extract(uint pos, uchar* p, uchar** uri);

		// HACERLO EN DOS FUNCIONES GENERICAS!!
		void encodeHT(uint code, uint len, uint *offset);
		void encodeHT(uint code, uint len, uchar *seq, uint *pos, uint *offset);
		char decodeHT(uint *pos, uint *offset);
		char decodeHT(uchar *seq, uint *pos, uint *offset);

		void decompressDelta(uint *pos, uint*offset);
		void decompressWord(uint *pos, uint*offset, uint ptr);

		/** Obtaints the length for the long common preffix of str1 and str2 */
		uint lcp(uchar* str1, uchar* str2, uint lstr1, uint lstr2);

		/** Locates the block for the given word and returns 'true' 
		    if it is the first word in the block */
		bool locateBlock(uchar *word, uint *block);

		/** Returns the word position in the block: 0 if it does
		    not exist. */
		uint locateWordInBlock(uint block, uchar *word);

		/** Returns the o-th word in 'block' **/
		uchar* locateWordInBlock(uint block, uint o);

		/** Buils the HuTucker Tree Coding **/
		static Node * buildHuTuckerTree(BitString_F *tree);
};

#endif  /* _FCHT_H_ */
