/* FCHTinterface.cpp
 * Copyright (C) 2011, Antonio Fariña and Miguel A Martinez-Prieto all rights reserved.
 * 
 * FCHTinterface.cpp: Handles a front-coding Hu-Tucker representation for a dictionary
 *   containing numberOfCanonicalWords. Provides functions to:
 *   create the compressed representation, save/load/free/size, and search (get)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 *
 */

/* Interface for the dictionary representation defined in a SEA 2011 paper:
 * Brisaboa, N.R; Cánovas, R.; Claude, F.; Martínez-Prieto, M.A.; Navarro, G.:
 * Compressed String Dictionaries. SEA 2011: 136-147.
 * 
 *   Miguel A Martinez-Prieto:  migumar2@infor.uva.es
 */



#include "FCHTinterface.h"

using namespace std;


//#ifdef COMPRESS_HUFFMAN_CANONICALS  

/** *******************************************************************
  * creating the compressed structure during construction of WCSA 
  * *******************************************************************/  	
int buildCanonicalsFCHT(void **HC,  uchar **words, uint numberOfCanonicalWords, size_t blocksize){
	FCHT *dicc = new FCHT(words,numberOfCanonicalWords, blocksize);;
	*HC = (void*)dicc;
}

int saveCanonicalsFCHT (void *HC, char *outfilename) {
	FCHT *dicc = (FCHT *) HC;
	ofstream out2(outfilename);
	dicc->save(out2);
	out2.close();		  
  return 0;
}

int loadCanonicalsFCHT (void **HC, char *filename) {
	ifstream rep(filename);
	
	FCHT *dicc = FCHT::load(rep);
	rep.close();			
	*HC = (void *) dicc;
	return 0;
}

int freeCanonicalsFCHT (void *HC) {
  FCHT * dicc = (FCHT*)HC;
  delete dicc;
  return 0;	
}

size_t sizeCanonicalsFCHT(void  *HC) {
	FCHT * dicc = (FCHT*)HC;
	return (size_t) dicc->getSize();  
 }

/** gets the ID of the word "s1" (if found) **/
size_t getIDFCHT(void *HC, uchar *s1, register ulong ws1){
	FCHT * dicc = (FCHT*)HC;
	size_t i =  dicc->locate(s1, (size_t) ws1);	
	return i-1;	
}

/** get the ID-th canonical word **/
uchar * getWordFCHT(void *HC, size_t id) {
	FCHT * dicc = (FCHT*)HC;
	uchar* word = dicc->extract(id+1);
	return (word);
}

/** DEBUGING ONLY****************/

void showCanonicalsFCHT(void *HC) {
	FCHT * dicc = (FCHT*)HC;
	size_t n= dicc->getLength();
	size_t i;
	fprintf(stderr,"\n Decoded canonicals:\n");
	for (i=0;i<n; i++) {
		uchar *dst;
		dst = getWordFCHT(HC, i);
		fprintf(stderr,"\n[Canonical word[%6u]: %s]",i, dst);
		fprintf(stderr,"[Len:%d]", strlen((char *)dst)) ;
	}
	fflush(stderr);
}


//#endif
