/* FCHTinterface.h
 * Copyright (C) 2011, Antonio Fariña and Miguel A Martinez-Prieto all rights reserved.
 * 
 * FCHTinterface.h: Handles a front-coding Hu-Tucker representation for a dictionary
 *   containing numberOfCanonicalWords. Provides functions to:
 *   create the compressed representation, save/load/free/size, and search (get)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 *
 */

/* Interface for the dictionary representation defined in a SEA 2011 paper:
 * Brisaboa, N.R; Cánovas, R.; Claude, F.; Martínez-Prieto, M.A.; Navarro, G.:
 * Compressed String Dictionaries. SEA 2011: 136-147.
 * 
 *   Miguel A Martinez-Prieto:  migumar2@infor.uva.es
 */

#include <string>
#include <vector>
#include <sstream>
#include <fstream> 
#include <iostream>
#include <algorithm>
#include <stdio.h>

#include "FCHT.h"

using namespace std;


/** *******************************************************************
  * creating the tCanonicalsHuff structure during construction of WCSA 
  * *******************************************************************/  	
int buildCanonicalsFCHT(void **HT,  uchar **words, uint numberOfCanonicalWords, size_t blocksize);     
int saveCanonicalsFCHT (void *HC, char *filename);

int loadCanonicalsFCHT (void **HC, char *filename);

int freeCanonicalsFCHT (void *HC);

size_t sizeCanonicalsFCHT(void  *HC);

/** gets the ID of the word "s1" (if found) **/
size_t getIDFCHT(void *HC, uchar *s1, register ulong ws1);

/** get the ID-th canonical word **/
uchar * getWordFCHT(void *HC, size_t id);

/** DEBUGING ONLY****************/
void showCanonicalsFCHT(void *HC);

//#endif
