/* tests.cpp
 * Copyright (C) 2010, Rodrigo Canovas & Miguel A. Martinez-Prieto
 * all rights reserved.
 *
 * Tests
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *
 * Contacting the author:
 *   Miguel A Martinez-Prieto:  migumar2@infor.uva.es
 */

#include <algorithm>
#include <fstream> 
#include <iostream>
#include <stdio.h>
#include <string>
#include <sys/time.h>
#include <vector>
#include <time.h>

#include "FCHTinterface.h"

#define SEC_TIME_DIVIDER  ((double) 1.0	)	 //1 sec = 1 sec
#define SEC_TIME_UNIT 	  "sec"

using namespace std;

bool 
sorting (string i, string j)
{ 
	return (i<j); 
}



void
buildMem(char *c_out)
{
	char *aux = new char[8];
	void *dicc;


	uint n=0;
	unsigned char **words = (unsigned char **) malloc ((1000000) * sizeof(unsigned char *));
	
	ifstream in(c_out);
	if(in.good())
	{
		string current;
		uint  lcurr;
		while (getline(in, current))
		{	
			lcurr = current.length();
			words[n] = (unsigned char *) malloc ((lcurr+1) * sizeof(unsigned char));
			uint z=0;
			for ( ;z<lcurr;z++)
			{
				words[n][z] = (unsigned char)current[z];
			}
			words[n][z]='\0';
			n++;
		}
		
		in.close();		
	}
	else
	{
		abort();
	}

	for (uint i=1; i<12; i++)
	{
		sprintf(aux,"%d", (int)pow(2,i));
		
		cout << "Building FHCT for sampling " << pow(2,i) << endl;
		
		//dicc = new FCHT(words,n, (size_t)pow(2,i));
		buildCanonicalsFCHT(&dicc, words, n ,(size_t)pow(2,i));
		
		ofstream out2((char*)(string(c_out)+string(".fcht.")+string(aux)).c_str());
		//dicc->save(out2);
		saveCanonicalsFCHT(dicc,(char*)(string(c_out)+string(".fcht.")+string(aux)).c_str());
		out2.close();
		//delete dicc;
		freeCanonicalsFCHT(dicc);

	}
	
	for (uint i=0;i<n;i++) free(words[i]);
	free(words);
	delete [] aux;
}


void
build(char *c_out)
{
	buildMem(c_out);
}



void
queryPatterns(char *c_index)
{
	ifstream index(c_index);	
	// PON AQUI EL FICHERO DE PATRONES CON LAS PALABRAS!!!
	ifstream patterns((char*)"clueweb09.patternsG");
	
	if (index.good() )
	{
		if ( patterns.good())
		{	
			//FCHT *dicc = FCHT::load(index);
			index.close();		
			void *dicc;
			loadCanonicalsFCHT(&dicc,c_index);
			
			struct timeval tv1,tv2;
			gettimeofday(&tv1,NULL);
			uint n = 0;
			string p;
			
			while (getline(patterns, p))
			{	
				//uint i = dicc->locate((uchar*)p.c_str(), p.length());
				uint i = getIDFCHT(dicc,(uchar*)p.c_str(), p.length());
				uchar *w = getWordFCHT(dicc,(size_t)i);
				printf("\n pattern: %20s has *id = %3u * --> getword(%3u) = %s",p.c_str(),i, i, w);
				
				n++;
			}
			printf("\n");
			
			gettimeofday(&tv2,NULL);
			patterns.close();
			
			printf("Tiempo total (%d ops): %ld en us\n",n, (tv2.tv_sec-tv1.tv_sec)*1000000+(tv2.tv_usec-tv1.tv_usec));
			printf("Tiempo por op: %f en us\n\n",((tv2.tv_sec-tv1.tv_sec)*1000000+(tv2.tv_usec-tv1.tv_usec))*1.0/(n*1.0));
			printf("Size of structure = %lu bytes\n",(unsigned long) sizeCanonicalsFCHT(dicc));
			//delete dicc;
			freeCanonicalsFCHT(dicc);
		}
		else
		{
			cout << endl << "   <ERROR> Check patterns file" << endl;
		}
	}
	else
	{
		cout << endl << "   <ERROR> Check index file" << endl;
	}


}
	
void
queryIDs(char *c_index)
{

	ifstream index(c_index);
	// PON AQUI EL FICHERO DE PATRONES CON LOS IDs!!!
	ifstream ids("clueweb09.patternsG.ids");
	
	if (index.good() && ids.good())
	{	
		//FCHT *dicc = FCHT::load(index);
		index.close();
		
		void *dicc;
		loadCanonicalsFCHT(&dicc,c_index);

		struct timeval tv1,tv2;
		gettimeofday(&tv1,NULL);
		uint n = 0;
		uint p;

		while (!ids.eof())
		{	
			p = loadValue<uint>(ids);
			//uchar* word = dicc->extract(p);
			uchar *word = getWordFCHT(dicc,(size_t)p-1);
			printf("\n id %3u belongs to pattern: %2s ",p,word);

			n++;
		}
		
		gettimeofday(&tv2,NULL);
		ids.close();
		
		printf("Tiempo total: %ld en us\n",(tv2.tv_sec-tv1.tv_sec)*1000000+(tv2.tv_usec-tv1.tv_usec));
		printf("Tiempo por op: %f en us\n\n",((tv2.tv_sec-tv1.tv_sec)*1000000+(tv2.tv_usec-tv1.tv_usec))*1.0/(n*1.0));
		
		//delete dicc;
		freeCanonicalsFCHT(dicc);
	}
	else
	{
		cout << endl << "   <ERROR> Check index and IDs file" << endl;
	}
}
				
int 
main(int argc, char* argv[])
{
	if (argc == 3) 
	{			
		int choice = atoi(argv[1]);
		
		switch (choice)
		{
			case 1:
				build(argv[2]);
				break;
			case 2:
				queryPatterns(argv[2]);
				break;
			case 3:
				queryIDs(argv[2]);
				break;
			case 4:
				buildMem(argv[2]);
				break;
			default:
				cout << "Non supported choice" << endl;
		}
			
	}
	else
	{
		// Bad number of parameters
		cout << endl << "   <ERROR> Wrong number of parameters: check your syntax..." << endl;
	}
}

