
/* canonicalsHuff.h
 * Copyright (C) 2011, Antonio Fariña, all rights reserved.
 *
 * canonicalsHuff.h: Handling the array of canonical words (the vocabulary).
 *   compressed with canonical Huffman (char-based, bit-oriented)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#ifdef COMPRESS_HUFFMAN_CANONICALS_type
	typedef struct {
		uint nwords;    //number of canonical words.
		uint *words;   // kbitVector. 1 extra bucket to point to the end
		uint elemSize;  //the size (in bits) of each pointer.
		uint *zoneMemHuff;
		uint sizeZoneHuff;  //size in bits!!
		
		uint maxLenAsciiCanonical; //the max len in bytes of a canonical word :: temp zone !!!
		byte *decodeBuffer;	//a buffer for decoding canonicals.	
			
		THuff H;
	} tCanonicalsHuff;                // '\0' char is not needed at the end. --> len[i] = words[i+1] - words[i];
#endif	


#ifdef COMPRESS_HUFFMAN_CANONICALS  

/** *******************************************************************
  * creating the tCanonicalsHuff structure during construction of WCSA 
  * *******************************************************************/  	
int buildHuffCanonicals(tCanonicalsHuff *HC,  byte **words, uint numberOfCanonicalWords); 

int saveCanonicals (tCanonicalsHuff *HC, int file);

int loadCanonicals (tCanonicalsHuff *HC, int file);

int freeCanonicals (tCanonicalsHuff *HC);

int sizeCanonicals(tCanonicalsHuff *HC);

//permits to compare 2 strings:
//  s1 in plain form. it contains ws1 and does not end in '\0'
//  s2 in compressed form. (it does not end in '\0')
int strcompHuff(tCanonicalsHuff *HC, uint i, const byte *s1, register ulong ws1);


/** DEBUGING ONLY****************/
void showCanonicals(tCanonicalsHuff *HC);

/** lenVariant 
	returns the length of a canonical word (in characters)
	*/
int lenCanonical(tCanonicalsHuff *HC, uint i);

/** asciiVariant 
	returns a pointer to a decoded variant (ended in '\0')
	*/
byte * asciiCanonical(tCanonicalsHuff *HC, uint i);


/** asciiVariant 
	returns a pointer to a decoded variant (ended in '\0')
	*/
byte * asciiCanonicalLen(tCanonicalsHuff *HC, uint i, uint *length);


#endif
