
/* filterWord.c
 * Copyright (C) 2011, Antonio Fariña, all rights reserved.
 *
 * filterWord.c: Obtains a canonical word associated to a given word.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


/*------------------------------------------------------------------
 getCanonical():: Obtains the canonical word of a given "word-variant".
 Uncomment the option you want for normalization: 
  	default choice: REMOVE_STOPWORDS_BEFORE_STEMMING
 ----------------------------------------------------------------- */


#include "filterWord.h"



/* uncomment your choice here: */

#define REMOVE_STOPWORDS_BEFORE_STEMMING
//#define APPLY_STEMMING_BEFORE_REMOVING_STOPWORDS
//#define NO_STEMMING

/* end: uncomment your choice here: */



#ifdef REMOVE_STOPWORDS_BEFORE_STEMMING

	int getCanonical(tCanonical *canonical,  byte *word, uint size, stt_hash stopWordsHash) {
		register byte *src, *dst;
		unsigned long i,addrInTH;
		
		dst= canonical->word; 
		src=word;

		for (i=0;i<size;i++) {*dst++ = _toLow[*src++];} //lowercase
		*dst='\0';

		i = stinHashTable(stopWordsHash, canonical->word,  size, &addrInTH);	// is it a stopword? 
		if (i) {
			canonical->type =IS_STOPWORD;
			return 0;
		}
		
		//non-stop-word
		size = stem((char *)canonical->word,0,size-1);
		canonical->word[size+1]=0;		
		canonical->size =size+1;  //?? check!!! esto es pq añade el '\0'
		canonical->type =IS_NORMALWORD;

		//	fprintf(stderr,"\n stemmed word is [%s],%d",canonical->word,canonical->size);fflush(stderr);
	return 0;
	}

#endif


#ifdef APPLY_STEMMING_BEFORE_REMOVING_STOPWORDS

	int getCanonical(tCanonical *canonical,  byte *word, uint size, stt_hash stopWordsHash) {
		register byte *src, *dst;
		unsigned long i,addrInTH;
		
		dst= canonical->word; 
		src=word;

		for (i=0;i<size;i++) {*dst++ = _toLow[*src++];} //lowercase
		*dst='\0';
		
		size = stem((char *)canonical->word,0,size-1);
		canonical->word[size+1]=0;
		
			
		canonical->size =size+1;  //?? check!!! esto es pq añade el '\0'
		
		i = stinHashTable(stopWordsHash,canonical->word, canonical->size, &addrInTH);	// is it a stopword?
		if (!i) canonical->type =IS_NORMALWORD;
		else	canonical->type =IS_STOPWORD;
		
	   //	fprintf(stderr,"\n stemmed word is [%s],%d",canonical->word,canonical->size);fflush(stderr);
	   
	return 0;   
	}
	
#endif


#ifdef NO_STEMMING

	int getCanonical(tCanonical *canonical,  byte *word, uint size, stt_hash stopWordsHash) {
		register byte *src, *dst;
		unsigned long i,addrInTH;
		
		dst= canonical->word; 
		src=word;

		for (i=0;i<size;i++) {*dst++ = _toLow[*src++];} //lowercase
		*dst='\0';
		canonical->size =size;
		
		i = stinHashTable(stopWordsHash,canonical->word, canonical->size, &addrInTH);	// is it a stopword?
		if (!i) canonical->type =IS_NORMALWORD;
		else	canonical->type =IS_STOPWORD;
	return 0;
	}
	
#endif
