
/* huffVariants.c
 * Copyright (C) 2011, Antonio Fariña, all rights reserved.
 *
 * huffVariants.c: Module with the functions needed to handle the variants
 *   structure.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */




//Compresses the stream "Fst|Num" used to store the Huffman Trees for each variant

/*
 	typedef struct {
 		uint *Dirb;    //Directory: offsets into zoneMem			
 		uint dirbSize; //in bits
 		uint dirbElemSize; // kbits (offset into zoneMem)
 		
 		uint *zoneMem;  // [Fst|Num][Fst|Num]...
 		uint zoneMemSize; //in bits
 		uint bucketSize; //in bits	
 		uint fstSize; //in bits	
 		uint numsize; //in bits	
 	} tHuffVariants;
*/


/** ************************************************************
  * creating the tVariants structure during construction of WCSA 
  * ************************************************************/  

#ifndef SEARCHERFACADEWITHMAIN  //in searchMode posInHT does not exist

/** compute max fst and max num */
void computeMaxNumDepth(uint *maxNum, uint *maxFst, uint *maxdepth,
		tposInHT *posInHT, uint numberOfCanonicalWords) {
			
	uint i,j, mempos, number,depth, posZone;

	uint maxN=0;
	uint maxF=0;
	uint maxD=0;						
	for(i = 0; i < numberOfCanonicalWords; i++) {
		depth =posInHT[i].huffman.depth;
		if (depth > maxD) maxD = depth;
		/**DEPTH*/
		//mempos+=depth +1;
		//for (j=0;j<(depth+1);j++) {
		for (j=0;j<(depth);j++) {
			 if (posInHT[i].huffman.fst[j]> maxF) maxF = posInHT[i].huffman.fst[j];
			 if (posInHT[i].huffman.num[j]> maxN) maxN = posInHT[i].huffman.num[j];
		}
	}	
	*maxNum = maxN;
	*maxFst = maxF;
	*maxdepth = maxD;
}

/*
 	typedef struct {
 		uint *Dirb;    //Directory: offsets into zoneMem			
 		uint dirbSize; //in bits
 		uint dirbElemSize; // kbits (offset into zoneMem)
 		
 		uint *zoneMem;  // [Fst|Num][Fst|Num]...
 		uint zoneMemSize; //in bits
 		uint sizeBuckbits; //in bits	
 		uint sizeFstbits; //in bits	
 		uint sizeNumbits; //in bits	
 	} tHuffVariants;
*/


/** build the structure */
int buildHuffVariants(tHuffVariants *HV, tposInHT *posInHT, 
		uint numberOfCanonicalWords, uint totallenFstAndNum) {

	uint i,j, mempos, number,depth, posDir,posZone;
	uint maxNum, maxFst, maxDepth;
	uint tmpValue;

	uint dirbElemSize, DirbSize; //directory
	uint zoneMemSize, sizeNumbits, sizeFstbits, sizeBuckbits; //for zonemem

	/** the maxFst, maxNum --> to obtain the number of bits needed per sample */
	computeMaxNumDepth(&maxNum, &maxFst, &maxDepth, posInHT,numberOfCanonicalWords);
	//fprintf(stderr,"\n maxnum = %d, maxfst = %d, maxdepth = %d",maxNum,maxFst,maxDepth);
	sizeNumbits = bits(maxNum);	sizeFstbits = bits(maxFst);
	sizeBuckbits = sizeNumbits + sizeFstbits;
	
	/** the sizes of the bit Vectors */
	zoneMemSize = totallenFstAndNum * sizeBuckbits;	
	dirbElemSize= bits(zoneMemSize); //to point to the end of zoneMemSize (in Dirb[numberOfCanonicalWords+1])
	DirbSize = (numberOfCanonicalWords+1)*dirbElemSize;   /**DEPTH2*/


	/** initializing Dirb */
	HV->dirbSize = DirbSize;
	HV->dirbElemSize = dirbElemSize;	
	HV->Dirb = (uint *) malloc ( ((HV->dirbSize+W-1)/W) * sizeof (uint) );
	HV->Dirb[((HV->dirbSize+W-1)/W) -1] = 0000;/**avoids valgrind to blame */

	HV->zoneMemSize = zoneMemSize;  //for num and fst
	HV->sizeBuckbits = sizeBuckbits;
	HV->sizeFstbits = sizeFstbits;
	HV->sizeNumbits = sizeNumbits;
	HV->zoneMem = (uint *) malloc ( ((HV->zoneMemSize + W -1)/W) * sizeof(uint) );	
	HV->zoneMem[((HV->zoneMemSize + W -1)/W) -1]= 0000;/**avoids valgrind to blame */


// fprintf(stderr,"\nDIR: dirbSize = %d, dirbElemSize = %d",HV->dirbSize,HV->dirbElemSize);fflush(stderr);
// fprintf(stderr,"\nOFF: zoneMemSize = %d, sizeBuckbits = %d",HV->zoneMemSize,HV->sizeBuckbits);fflush(stderr);
// fprintf(stderr,"\n     sizeFstbits = %d, sizeNumbits = %d",HV->sizeFstbits,HV->sizeNumbits);fflush(stderr);
// fprintf(stderr,"\nOTH:     numberOfCanonicalWords = %d, totallenFstAndNum = %d",numberOfCanonicalWords,totallenFstAndNum);fflush(stderr);
//

	/**setting the offsets in HV->Dirb, and copying fst & num arrays to zoneMemSize*/
	posZone=0;
	mempos=0;	
	posDir=0;					
	for(i = 0; i < numberOfCanonicalWords; i++) {
		//HV->Dirb[i*elemSizeDir] = mempos;
		bitwrite (HV->Dirb, posDir, dirbElemSize, mempos);	posDir += dirbElemSize;

		depth =posInHT[i].huffman.depth;
		/**DEPTH*/
		//mempos+=depth +1;
		//for (j=0;j<(depth+1);j++) {
		mempos+=depth* sizeBuckbits; 
		for (j=0;j<(depth);j++) {
			tmpValue = posInHT[i].huffman.fst[j];
			//HV->zoneMem[posZone] = tmpValue;
			bitwrite (HV->zoneMem, posZone, sizeFstbits, tmpValue); posZone+=sizeFstbits;
			
			tmpValue = posInHT[i].huffman.num[j];
			//HV->zoneNumFst[posZone] = tmpValue;
			bitwrite (HV->zoneMem, posZone, sizeNumbits, tmpValue);	posZone+=sizeNumbits;
		}
		//so mempos has actually the same value of posZone in this moment
	}		
	//set mempos also in the last position to avoid the need of storing "depth" !!
	//HV->Dirb[i*elemSizeDir]=mempos;  //such that, i == numberOfCanonicalWords
	bitwrite (HV->Dirb, posDir, dirbElemSize, mempos);	
	
	
	/**checking dir ok. */
//	uint offbucket,prev;
//	prev=0;
//	depth;
//	for (i=0;i<numberOfCanonicalWords;i++) {
//	 offbucket = bitread (HV->Dirb, i*(HV->dirbElemSize), (HV->dirbElemSize)); 
//	 depth = posInHT[i].huffman.depth;
//	 
//	 //fprintf(stderr,"\n ERRor i=%d, prev=%d, buck=%d !!",i,prev,offbucket);
//	 
//	 if (prev !=offbucket) {fprintf(stderr,"\n ERRor i=%d, prev=%d, buck=%d !!",i,prev,offbucket);exit(0);}
//	 	prev+= (HV->sizeBuckbits *depth);
//	}
	
	
}
#endif



int saveHuffVariants(tHuffVariants *HV, int file) {	
	//the array of offsets for each canonical word, to the beggining of [fst|num] buckets.
	write(file,&(HV->dirbSize), sizeof(uint));
	write(file,&(HV->dirbElemSize), sizeof(uint));
	write(file, (HV->Dirb),  ((HV->dirbSize+W-1)/W) * sizeof (uint) );

	//the buckets [fst0|num0][fst1|num1]... for the huffman trees of each canonical word.
	write(file,&(HV->zoneMemSize), sizeof(uint));
	write(file,&(HV->sizeBuckbits), sizeof(uint));
	write(file,&(HV->sizeFstbits), sizeof(uint));
	write(file,&(HV->sizeNumbits), sizeof(uint));
	write(file, (HV->zoneMem), ( ((HV->zoneMemSize + W -1)/W) * sizeof(uint) ));
 		
}

int loadHuffVariants(tHuffVariants *HV, int file) {	
	//the array of offsets for each canonical word, to the beggining of [fst|num] buckets.
	read(file,&(HV->dirbSize), sizeof(uint));
	read(file,&(HV->dirbElemSize), sizeof(uint));
	HV->Dirb = (uint *) malloc ( ((HV->dirbSize+W-1)/W) * sizeof (uint) );
	read(file, (HV->Dirb), ( ((HV->dirbSize+W-1)/W) * sizeof (uint)) );

	//the buckets [fst0|num0][fst1|num1]... for the huffman trees of each canonical word.
	read(file,&(HV->zoneMemSize), sizeof(uint));
	read(file,&(HV->sizeBuckbits), sizeof(uint));
	read(file,&(HV->sizeFstbits), sizeof(uint));
	read(file,&(HV->sizeNumbits), sizeof(uint));
	HV->zoneMem = (uint *) malloc ( ((HV->zoneMemSize + W -1)/W) * sizeof(uint) );	
	read(file, (HV->zoneMem), ( ((HV->zoneMemSize + W -1)/W) * sizeof(uint) ));	
}


int freeHuffVariants(tHuffVariants *HV) {
	free(HV->Dirb);
	free(HV->zoneMem);
	return 0;
}

int sizeHuffVariants(tHuffVariants *HV) {
	return  ( (((HV->dirbSize)+W-1)/W) * sizeof (uint)) + 
			( ((HV->zoneMemSize + W-1)/W) * sizeof(uint));
}


/** #define decodeHuffDecMacroVariantWordPos(fst, offbucket, symb, stream, ptr, depth) ==>
		==> SEE huffDec.h
*/


uint decodeHuffDecFuncVariantWordPos2bits(tHuffVariants HV, uint *symb, uint *stream, uint ptr, uint idCanonical) 
   { uint pos; 
     register uint d; 
     uint sizeBuckbits= HV.sizeBuckbits; 
     uint dirbElemSize = HV.dirbElemSize; 
     uint sizeFstbits = HV.sizeFstbits; 
     uint sizeNumbits = HV.sizeNumbits; 
     uint *dirb = HV.Dirb; 
     uint *zonefstnum = HV.zoneMem; 
	 register uint offbucket; 
	 offbucket = bitread (dirb, idCanonical*dirbElemSize, dirbElemSize); 
	 register uint depth; 
	 depth = bitread (dirb, (idCanonical+1)*dirbElemSize, dirbElemSize); 
	 depth = (depth - offbucket)/sizeBuckbits; 
     pos = 0; 
     d = 0; 
     register uint currfst; 
//     fprintf(stderr,"\n offbucket = %d, sizeBuckbits = %d, sizeFstbits = %d",offbucket,sizeBuckbits,sizeFstbits);fflush(stderr);
//     fprintf(stderr,"\n (offbucket + d*sizeBuckbits) = %d, sizeZone =%d",(offbucket + d*sizeBuckbits),HV.zoneMemSize);fflush(stderr);
     currfst = bitread (zonefstnum, (offbucket + d*sizeBuckbits), sizeFstbits); 
     while ((pos < currfst)  ) 
        { pos = (pos << 1) | bitget(stream,ptr); 
          ptr++; d++; 
          if (d<depth) 
          	currfst = bitread (zonefstnum, (offbucket + d*sizeBuckbits), sizeFstbits); 
          else break; 
        } 
    if (depth==d) 
     	*symb=pos; 
    else { 
    	uint currNum; 
    	currNum = bitread (zonefstnum, (offbucket + d*sizeBuckbits+ sizeFstbits), sizeNumbits); 
     	*symb = currNum + pos - currfst; 
    } 
     return ptr;
  }
          
//FALLA !!!!! ////
#define decodeHuffDecMacroVariantWordPos2bits(HV, symb, stream, ptr, idCanonical) \
   { uint pos; \
     register uint d; \
     uint sizeBuckbits= HV.sizeBuckbits; \
     uint dirbElemSize = HV.dirbElemSize; \
     uint sizeFstbits = HV.sizeFstbits; \
     uint sizeNumbits = HV.sizeNumbits; \
     uint *dirb = HV.Dirb; \
     uint *zonefstnum = HV.zoneMem; \
	 register uint offbucket; \
	 offbucket = bitread (dirb, idCanonical*dirbElemSize, dirbElemSize); \
	 register uint depth; \
	 depth = bitread (dirb, (idCanonical+1)*dirbElemSize, dirbElemSize); \
	 depth = (depth - offbucket)/sizeBuckbits; \
     pos = 0; \
     d = 0; \
     register uint currfst; \
     currfst = bitread (zonefstnum, (offbucket + d*sizeBuckbits), sizeFstbits); \
     while ((pos < currfst)  ) \
        { pos = (pos << 1) | bitget(stream,ptr); \
          ptr++; d++; \
          if (d<depth) \
          	currfst = bitread (zonefstnum, (offbucket + d*sizeBuckbits), sizeFstbits); \
          else break; \
        } \
    if (depth==d) \
     	symb=pos; \
    else { \
    	uint currNum; \
    	currNum = bitread (zonefstnum, (offbucket + d*sizeBuckbits+ sizeFstbits), sizeNumbits); \
     	symb = currNum + pos - currfst; \
    } \
  }          
