
/* separators.c
 * Copyright (C) 2011, Antonio Fariña, all rights reserved.
 *
 * separators.c: Module with the functions required to handle the separators
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

/*int buildSeparatorsb(tSeparatorsB *separators, tZoneMem *zoneAscii) {
return -1;
}
*/
/** ************************************************************
  * creating the tVariants structure during construction of WCSA 
  * ************************************************************/  

#ifndef SEARCHERFACADEWITHMAIN  //in searchMode posInHT does not exist
int buildSeparators(tSeparators2 *separators, tZoneMem *zoneAscii) {
	register uint i, j;
	byte *src;
	uint len;
	uint maxLenSep=0;

	uint numSeparators = separators->count;
	register uint lenAscii = zoneAscii->size;
	
	/** creating Huffman over ZoneMem (the text with all the variants) */
	uint freqs[256];
	for (i=0;i<256;i++) freqs[i]=0;
	src=zoneAscii->zone;
	i=0;
	while (i<lenAscii) {
		len=0;
		while (*src) {freqs[*src]++;src++; len++;}  
		//freqs[*src]++; //	freqs[0] =0; //the '\0' is no longer encoded;
		src++;
		if (maxLenSep < len) maxLenSep = len;
		i+=len+1; //jumping the '\0'
	}

	separators->maxLenSep = maxLenSep;
	separators->decodeBuffer = (byte*) malloc((maxLenSep +1)*sizeof(byte));

	/** creating the char-oriented Huffman tree */
	separators->H = createHuff (freqs,255,UNSORTED);

	/** allocating the memory needed for all the separators (Huffman); */
	separators->zoneMemHuff = (uint *) malloc (((separators->H.total + W -1)/W) * sizeof(uint));	
	separators->zoneMemHuff[((separators->H.total + W -1)/W)-1] = 0000; /**avoids valgrind to blame */
	separators->sizeZoneHuff = separators->H.total;


//	fprintf(stderr,"\n len ascii  = %d ",lenAscii);
//	fprintf(stderr,"\n SEP COUNT  = %d ",separators->count);
//	fprintf(stderr,"\n MAX LEN SEP = %d ",separators->maxLenSep);
//	fprintf(stderr,"\n SIZE ZONE HUF = %d ",separators->sizeZoneHuff);
//	fflush(stderr);


	
	/** Setting the pointers in Dirb and Subb */
	{
		byte *end;
		uint posDir, offDir;
		uint ptr,ptrprev;
		
		uint elemSizeDir = bits(separators->sizeZoneHuff+1);	//number of bits per sample
		uint DirbSize = (numSeparators+1)*elemSizeDir;  //numberOfSeparators;
				//1 extra element is kept at the end to obtain lenHuffman[i] = dir[i+1]-dir[i]; instead of the '\0'
				//so decoding is done as while (ptr < dir[i+1]) ...
	
		/** initializing Dirb */
		separators->Dirb =(uint *) malloc ( ((DirbSize+W-1)/W) * sizeof (uint));
		separators->Dirb[((DirbSize+W-1)/W) -1] = 0000; /**avoids valgrind to blame */
		separators->dirbSize = DirbSize;  		//size in bits of Dirb
		separators->dirbElemSize = elemSizeDir;	//size in bits of the elements of Dirb.

		/** generating  the compressed zoneMemHuff */
		posDir =0; offDir=0;
		ptr=ptrprev=0;

		src=zoneAscii->zone;
		end = src+lenAscii;
		while (src < end) {
			//Dir[posDir++]= offDir;
			bitwrite (separators->Dirb, posDir, elemSizeDir, offDir);	posDir += elemSizeDir;

			ptrprev=ptr;		
			while (*src) {
				ptr=encodeHuff(separators->H, *src, separators->zoneMemHuff, ptr);			
				src++; 
			}  
			src++; //the '\0' is not kept;
			len=ptr-ptrprev;
			offDir +=len;
		}
		//extra element
		bitwrite (separators->Dirb, posDir, elemSizeDir, offDir);	//the last one (extra element to mark the end).
	}
	/** set the Huffman tree to permit decoding */
	prepareToDecode(&(separators->H));


//	{/**checks decoding the compressed stream and comparing agains the source text (skips '\0')*/
//		uint i,decValue;		
//		uint ptr,ptrend;		
//		ptr=0; ptrend = separators->dirbSize;
//
//		THuff *huff = &(separators->H);
//		uint *stream=separators->zoneMemHuff;
//		byte *zoneascii = zoneAscii->zone;
//		
//		i=0;
//		while (ptr < ptrend) {
//			decodeNormalHuffMacro(huff,decValue,stream,ptr); 
//			if (zoneascii[i]==0) i++;
//			if (zoneascii[i]!= ((byte) decValue) ) {printf("\nSON DISTINTOS !!");fflush(stdout);exit(0);}
//			i++;
//		}		
//
//	}
}
#endif 


int saveSeparators (tSeparators2 *separators, int file) {

	/** the max length of a variant */
	write(file, &(separators->maxLenSep), sizeof(uint));	

	/** compressed variants */
	write(file, &(separators->sizeZoneHuff), sizeof(uint));
	write(file, (separators->zoneMemHuff), sizeof(uint) * ((separators->sizeZoneHuff + W-1)/W) );

	/** the huffman tree */
	THuff H = separators->H;
	write(file, &H.max, sizeof(int));
	write(file, &H.lim, sizeof(int));
	write(file, &H.depth, sizeof(int));
	write(file, H.s.symb, (H.lim+1)*sizeof(int));	
	write(file, H.num, (H.depth+1)*sizeof(int));
	write(file, H.fst, (H.depth+1)*sizeof(int));

	/** dirb */
	write(file, &(separators->dirbSize), sizeof(uint));
	write(file, &(separators->dirbElemSize), sizeof(uint));
	write(file, separators->Dirb, ((separators->dirbSize + W-1)/W) * sizeof(uint));
}

int loadSeparators (tSeparators2 *separators, int file) {

	/** the max length of a variant */
	read(file, &(separators->maxLenSep), sizeof(uint));	
	separators->decodeBuffer = (byte*) malloc((separators->maxLenSep +1)*sizeof(byte)); //tmp buffer

	/** compressed variants */
	read(file, &(separators->sizeZoneHuff), sizeof(uint));
	separators->zoneMemHuff = (uint *) malloc (sizeof(uint) * ((separators->sizeZoneHuff + W-1)/W));
	read(file, (separators->zoneMemHuff), sizeof(uint) * ((separators->sizeZoneHuff + W-1)/W) );

	/** the huffman tree */
	THuff H;
	read(file, &H.max, sizeof(int));
	read(file, &H.lim, sizeof(int));
	read(file, &H.depth, sizeof(int));
	H.s.symb = (unsigned int *) malloc((H.lim+1)*sizeof(int));
	H.num = (unsigned int *) malloc((H.depth+1)*sizeof(int));	
	H.fst = (unsigned int *) malloc((H.depth+1)*sizeof(int));	

	read(file, H.s.symb, (H.lim+1)*sizeof(int));	
	read(file, H.num, (H.depth+1)*sizeof(int));
	read(file, H.fst, (H.depth+1)*sizeof(int));	
	separators->H = H;

	/** Dirb */
	read(file, &(separators->dirbSize), sizeof(uint));
	read(file, &(separators->dirbElemSize), sizeof(uint));
	separators->Dirb =(uint *) malloc ( (((separators->dirbSize)+W-1)/W) * sizeof (uint));	
	read(file, separators->Dirb, ((separators->dirbSize + W-1)/W) * sizeof(uint));
}

int freeSeparators (tSeparators2 *separators) {
	free(separators->zoneMemHuff);
	free(separators->decodeBuffer);
	free(separators->Dirb);
	freeHuff(separators->H);
	return 0;
}

int sizeSeparators(tSeparators2 *separators) {
	return  ( (((separators->dirbSize)+W-1)/W) * sizeof (uint)) + 
			( ((separators->sizeZoneHuff + W-1)/W) * sizeof(uint)) +
			( ((separators->maxLenSep) +1) * sizeof(byte)) + 
			sizeHuff (separators->H);
}



/** lenSeparator
	returns the length of a separator (in characters)
	*/
int lenSeparator(tSeparators2 *separators, uint i) {
	register uint len;
	uint dirPos;	
	register uint ptr,ptrEnd;
		
	dirPos = i * separators->dirbElemSize;   // Dir[i*esize]
	ptr  = bitread ( separators->Dirb, dirPos, separators->dirbElemSize); //starting decode Pos
	dirPos +=  separators->dirbElemSize;	// Dir[(i+1)*esize]
	ptrEnd  = bitread (separators->Dirb, dirPos, separators->dirbElemSize); //ending decode Pos
			//	mybitread(absPos,variants->Subb,posSub,variants->subbAbsSize);	

	{	//Decoding until the begining of the next separator.
		register uint decValue;
		THuff *huff = &(separators->H);
		uint *stream=separators->zoneMemHuff;
		len=0;
		while (ptr < ptrEnd) {
		//ptr = decodeHuff(huff,&variant,variants->zoneMemHuff,ptr); 
			decodeNormalHuffMacro(huff,decValue,stream,ptr); 
			len++; //a decoded char.
		}		
	}
	return (len);
}



/** asciiSeparator 
	returns a pointer to a decoded separator (ended in '\0')
	*/
byte * asciiSeparator(tSeparators2 *separators, uint i) {
	byte *dst;
	register uint len;
	uint dirPos;	
	register uint ptr,ptrEnd;
		
	dirPos = i * separators->dirbElemSize;   // Dir[i*esize]
	ptr  = bitread ( separators->Dirb, dirPos, separators->dirbElemSize); //starting decode Pos
	dirPos +=  separators->dirbElemSize;	// Dir[(i+1)*esize]
	ptrEnd  = bitread (separators->Dirb, dirPos, separators->dirbElemSize); //ending decode Pos
			//	mybitread(absPos,variants->Subb,posSub,variants->subbAbsSize);	
	dst = separators->decodeBuffer;
	{	//Decoding until the begining of the next separator.
		register uint decValue;
		THuff *huff = &(separators->H);
		uint *stream=separators->zoneMemHuff;
		len=0;
		while (ptr < ptrEnd) {
		//ptr = decodeHuff(huff,&variant,variants->zoneMemHuff,ptr); 
			decodeNormalHuffMacro(huff,decValue,stream,ptr); 
			*dst=(byte)decValue; dst++;
			len++; //a decoded char.
		}		
		*dst='\0';
	}
	return ((byte*) (separators->decodeBuffer));
}

/** asciiSeparator 
	returns a pointer to a decoded separator (ended in '\0')
	*/
byte * asciiSeparatorLen(tSeparators2 *separators, uint i, uint *length) {
	byte *dst;
	register uint len;
	uint dirPos;	
	register uint ptr,ptrEnd;
		
	dirPos = i * separators->dirbElemSize;   // Dir[i*esize]
	ptr  = bitread ( separators->Dirb, dirPos, separators->dirbElemSize); //starting decode Pos
	dirPos +=  separators->dirbElemSize;	// Dir[(i+1)*esize]
	ptrEnd  = bitread (separators->Dirb, dirPos, separators->dirbElemSize); //ending decode Pos
			//	mybitread(absPos,variants->Subb,posSub,variants->subbAbsSize);	
	dst = separators->decodeBuffer;
	{	//Decoding until the begining of the next separator.
		register uint decValue;
		THuff *huff = &(separators->H);
		uint *stream=separators->zoneMemHuff;
		len=0;
		while (ptr < ptrEnd) {
		//ptr = decodeHuff(huff,&variant,variants->zoneMemHuff,ptr); 
			decodeNormalHuffMacro(huff,decValue,stream,ptr); 
			*dst=(byte)decValue; dst++;
			len++; //a decoded char.
		}		
		*dst='\0';
	}
	*length=len;
	return ((byte*) (separators->decodeBuffer));
}

/** asciiSeparatorCopyBytes
	decodes a Separator and copies the first "number" bytes to decodedSep
	*/
void asciiSeparatorCopyBytes(byte *decodedSep, tSeparators2 *separators, uint i, register uint number){
	byte *dst;
	register uint len;
	uint dirPos;	
	register uint ptr;
		
	dirPos = i * separators->dirbElemSize;   // Dir[i*esize]
	ptr  = bitread ( separators->Dirb, dirPos, separators->dirbElemSize); //starting decode Pos
			//	mybitread(absPos,variants->Subb,posSub,variants->subbAbsSize);	
	dst = decodedSep;
	{	//Decoding until the begining of the next separator.
		register uint decValue;
		register uint num;
		THuff *huff = &(separators->H);
		uint *stream=separators->zoneMemHuff;
		len=0;
		for (num=0;num<number;num++) {
			decodeNormalHuffMacro(huff,decValue,stream,ptr); 
			//ptr = decodeHuff(huff,&variant,variants->zoneMemHuff,ptr); 
			*dst=(byte)decValue; dst++;
		}				
	}
}




/** DEBUGING ONLY****************/

//void showVariants2(tVariants *variants){
//	uint i;
// 	fprintf(stderr,"\n zoneMem:\n");
// 	for (i=0;i<120; i++) {
// 		fprintf(stderr,"%c",variants->zoneMem[i]);
// 	}
// 	fprintf(stderr,"\n");
// 	for (i=0;i<120; i++) {
// 		if (!(i%10)) fprintf(stderr,"1");
// 		else fprintf(stderr," ");
// 	}
