/* variants.c
 * Copyright (C) 2011, Antonio Fariña, all rights reserved.
 *
 * variants.c: Module with the functions required to handle the variants.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */



/** ************************************************************
  * creating the tVariants structure during construction of WCSA 
  * ************************************************************/  

#ifndef SEARCHERFACADEWITHMAIN  //in searchMode posInHT does not exist
int buildVariants(tVariants *variants, tposInHT *posInHT, uint numberOfCanonicalWords,
				  uint totallenVariants, uint numberOfDistinctValidWords) {
	uint maxJump=0;
	uint maxLenAsciiVariant =0;
	register uint i, j;
	uint number,len,jump;
	byte *src;

	/** creating Huffman over ZoneMem (the text with all the variants) */
	uint freqs[256];
	for (i=0;i<256;i++) freqs[i]=0;
	for (i=0;i<numberOfCanonicalWords;i++) {
		number = posInHT[i].number;
		for (j=0;j<number;j++) {
			src = posInHT[i].variants[j];		 //calculates the len of the variant
			len=0;
			while (*src) {freqs[*src]++;src++; len++;}  
			freqs[*src]++;
			if (maxLenAsciiVariant < len) maxLenAsciiVariant = len;
		}
	}
	variants->maxLenAsciiVariant = maxLenAsciiVariant;
	variants->decodeBuffer = (byte*) malloc((maxLenAsciiVariant +1)*sizeof(byte));

	variants->H = createHuff (freqs,255,UNSORTED);

	/** allocating the memory needed for all variants (Huffman); */
	variants->zoneMemHuff = (uint *) malloc ((variants->H.total + W -1)/W * sizeof(uint));
	variants->sizeZoneHuff = variants->H.total;

	/** Computing the max jump that can be needed for the 2nd, 3rd, etc variants of a word
		in order to know the number of bits that have to be reserved.
		!! And generating the compressed text in zoneMemHuff (it will be repeated later).
    */
	uint ptr=0;
	uint ptrprev=0;

	for (i=0;i<numberOfCanonicalWords;i++) {
		number = posInHT[i].number;
		//the first Variant (pos 0)
		j=0;
		src = posInHT[i].variants[j];		//calculates the len of the variant
		ptrprev=ptr;		
		while (*src) {
			ptr=encodeHuff(variants->H, *src, variants->zoneMemHuff, ptr);			
			src++; 
		}  
		ptr=encodeHuff(variants->H, *src, variants->zoneMemHuff, ptr); //includes the '\0' in the length
		len=ptr-ptrprev;					            
		jump=len; 	//initializes jump for the variants of current canonical word.

		for (j=1;j<number;j++) {			
			src = posInHT[i].variants[j];		 //calculates the len of the variant
			ptrprev=ptr;		
			while (*src) {
				ptr=encodeHuff(variants->H, *src, variants->zoneMemHuff, ptr);
				src++; 
			} 
			ptr=encodeHuff(variants->H, *src, variants->zoneMemHuff, ptr); //includes the '\0' in the length
			len=ptr-ptrprev;					            
			if (maxJump < jump) maxJump = jump; /**max Jump is only increased here, as the jump for words
											    with only one variant does not import... and the maxjump
												refers to the jump to the "begining of the last variant of a word".*/
			jump +=len;
		}		
	}

	/** Setting the pointers in Dirb and Subb */
	{
		byte *zoneMem;
		uint posDir, posSub, offDir,offSub;
		
		/** allocating the memory needed for all variants; */

		uint numbitsAbsValueSubb = bits(variants->sizeZoneHuff);  //length of absolute sample in Subb to Variants vector(compres)
		uint numbitsMaxJump = bits(maxJump);      //length of a jump (in bits).
		uint numJumpsInSub = numberOfDistinctValidWords - numberOfCanonicalWords; //number of relative jumps in Subb
		uint SubbSize = numJumpsInSub * numbitsMaxJump + numberOfCanonicalWords * numbitsAbsValueSubb; //number of bits in Subb
		uint elemSizeDir = bits(SubbSize);
		uint DirbSize = (numberOfCanonicalWords+1) * elemSizeDir; /*****@ +1 @*****/
	
		/** initializing Dirb */
				
		//variants->Dirb = create_kBitArray (numberOfCanonicalWords,elemSizeDir);
		variants->Dirb =(uint *) malloc ( ((DirbSize+W-1)/W) * sizeof (uint));
		variants->Dirb[((DirbSize+W-1)/W) -1] = 0000; /**avoids valgrind to blame */
		variants->dirbSize = DirbSize;  		//size in bits of Dirb
		variants->dirbElemSize = elemSizeDir;	//size in bits of the elements of Dirb.

		/** initializing Subb */
		variants->Subb =(uint *) malloc ( ((SubbSize+W-1)/W) * sizeof (uint));
		variants->Subb[((SubbSize+W-1)/W) -1] = 0000; /**avoids valgrind to blame */
		variants->subbSize = SubbSize;  //size in bits of Subb
		variants->subbAbsSize = numbitsAbsValueSubb ; // kbits (offset into zoneMem)
		variants->subbJumpSize = numbitsMaxJump; // kbits (bits to represent a jump)

		//clear the compressed zone	(the text is compressed again) !!
		i=0; while (i<ptr) { variants->zoneMemHuff[i/W]=0; i+=W;}
		
		/** generating again the compressed zoneMemHuff */

		ptr=ptrprev=0;
		posDir=0; posSub=0;
		offDir=0; offSub=0;

		
		for (i=0;i<numberOfCanonicalWords;i++) {
			number = posInHT[i].number;
			//Dir[posDir++]= offDir;		/** Dir entry */
			bitwrite (variants->Dirb, posDir, elemSizeDir, offDir);	posDir += elemSizeDir;
			
			{	/** the first word has an absolute offset*/
				//Sub[posSub++]=offSub;   /** Sub entry */			
				bitwrite (variants->Subb, posSub, numbitsAbsValueSubb, offSub);
				posSub+=numbitsAbsValueSubb;
				j=0;
				src = posInHT[i].variants[j];		 //copying the variant of the word
				ptrprev=ptr;		
				while (*src) {
					ptr=encodeHuff(variants->H, *src, variants->zoneMemHuff, ptr);			
					src++; 
				}  				
				ptr=encodeHuff(variants->H, *src, variants->zoneMemHuff, ptr); //includes the '\0' in the length
				len=ptr-ptrprev;
				offSub +=len;
				jump=len;
				offDir+=numbitsAbsValueSubb; // to point to the next empty-bit in Subb.
			}
			
			/** the next variants is kept as a jump + abs value */
			for (j=1;j<number;j++) {			
				//Sub[posSub++]=jump;   /** Sub entry */
				bitwrite (variants->Subb, posSub, numbitsMaxJump, jump);
				posSub+=numbitsMaxJump;		
				src = posInHT[i].variants[j];		 //calculates the len of the variant
				ptrprev=ptr;		
				while (*src) {
					ptr=encodeHuff(variants->H, *src, variants->zoneMemHuff, ptr);
					src++; 
				} 
				ptr=encodeHuff(variants->H, *src, variants->zoneMemHuff, ptr); //includes the '\0' in the length
				len=ptr-ptrprev;
				offSub +=len;
				jump +=len;
				offDir+=numbitsMaxJump; // to point to the next empty-bit in Subb.
			}		
		}		
		
		bitwrite (variants->Dirb, posDir, elemSizeDir, offDir);	/*****@ writes ptr to end of subb @*****/

	}



	/** set the Huffman tree to permit decoding */
	prepareToDecode(&(variants->H));
}
#endif 

int saveVariants (tVariants *variants, int file) {

	/** the max length of a variant */
	write(file, &(variants->maxLenAsciiVariant), sizeof(uint));	

	/** compressed variants */
	write(file, &(variants->sizeZoneHuff), sizeof(uint));
	write(file, (variants->zoneMemHuff), sizeof(uint) * ((variants->sizeZoneHuff + W-1)/W) );

	/** the huffman tree */
	THuff H = variants->H;
	write(file, &H.max, sizeof(int));
	write(file, &H.lim, sizeof(int));
	write(file, &H.depth, sizeof(int));
	write(file, H.s.symb, (H.lim+1)*sizeof(int));	
	write(file, H.num, (H.depth+1)*sizeof(int));
	write(file, H.fst, (H.depth+1)*sizeof(int));

	/** dirb */
	write(file, &(variants->dirbSize), sizeof(uint));
	write(file, &(variants->dirbElemSize), sizeof(uint));
	write(file, variants->Dirb, ((variants->dirbSize + W-1)/W) * sizeof(uint));

	/** Subb */
	write(file, &(variants->subbSize), sizeof(uint));
	write(file, &(variants->subbJumpSize), sizeof(uint));
	write(file, &(variants->subbAbsSize), sizeof(uint));
	write(file, variants->Subb, ((variants->subbSize + W-1)/W) * sizeof(uint));
}

int loadVariants (tVariants *variants, int file) {

	/** the max length of a variant */
	read(file, &(variants->maxLenAsciiVariant), sizeof(uint));	
	variants->decodeBuffer = (byte*) malloc((variants->maxLenAsciiVariant +1)*sizeof(byte)); //tmp buffer

	/** compressed variants */
	read(file, &(variants->sizeZoneHuff), sizeof(uint));
	variants->zoneMemHuff = (uint *) malloc (sizeof(uint) * ((variants->sizeZoneHuff + W-1)/W));
	read(file, (variants->zoneMemHuff), sizeof(uint) * ((variants->sizeZoneHuff + W-1)/W) );

	/** the huffman tree */
	THuff H;
	read(file, &H.max, sizeof(int));
	read(file, &H.lim, sizeof(int));
	read(file, &H.depth, sizeof(int));
	H.s.symb = (unsigned int *) malloc((H.lim+1)*sizeof(int));
	H.num = (unsigned int *) malloc((H.depth+1)*sizeof(int));	
	H.fst = (unsigned int *) malloc((H.depth+1)*sizeof(int));	

	read(file, H.s.symb, (H.lim+1)*sizeof(int));	
	read(file, H.num, (H.depth+1)*sizeof(int));
	read(file, H.fst, (H.depth+1)*sizeof(int));	
	variants->H = H;

	/** Dirb */

	read(file, &(variants->dirbSize), sizeof(uint));
	read(file, &(variants->dirbElemSize), sizeof(uint));
	variants->Dirb =(uint *) malloc ( (((variants->dirbSize)+W-1)/W) * sizeof (uint));	
	read(file, variants->Dirb, ((variants->dirbSize + W-1)/W) * sizeof(uint));

	/** Subb */
	read(file, &(variants->subbSize), sizeof(uint));
	read(file, &(variants->subbJumpSize), sizeof(uint));
	read(file, &(variants->subbAbsSize), sizeof(uint));
	variants->Subb =(uint *) malloc ( (((variants->subbSize)+W-1)/W) * sizeof (uint));	
	read(file, variants->Subb, ((variants->subbSize + W-1)/W) * sizeof(uint));	
	//printf("\n %d (bits) %d %d", variants->subbSize, variants->subbJumpSize, variants->subbAbsSize);
}

int freeVariants (tVariants *variants) {
	free(variants->zoneMemHuff);
	free(variants->decodeBuffer);
	free(variants->Dirb);
	free(variants->Subb);
	freeHuff(variants->H);
	return 0;
}

int sizeVariants(tVariants *variants) {
	return  ( (((variants->dirbSize)+W-1)/W) * sizeof (uint)) + 
			( (((variants->subbSize)+W-1)/W) * sizeof (uint)) + 
			( ((variants->sizeZoneHuff + W-1)/W) * sizeof(uint)) +
			( ((variants->maxLenAsciiVariant) +1) * sizeof(byte)) +
			sizeHuff (variants->H);
}



/** lenVariant 
	returns the length of a variant (in characters)
	*/
int lenVariant(tVariants *variants, uint i, uint j) {
	register uint len;
	uint posSub, absPos;	
	uint jump;
	uint postmp;
	uint returnValue;
		
	postmp = i * variants->dirbElemSize;
	posSub  = bitread ( variants->Dirb, postmp, variants->dirbElemSize);
	//	mybitread(absPos,variants->Subb,posSub,variants->subbAbsSize);
	absPos = bitread (variants->Subb,posSub,variants->subbAbsSize);
	jump=0;
	if (j!=0) {
		uint tmp;
		postmp =  (posSub+ (variants->subbAbsSize) + (variants->subbJumpSize * (j-1)));
		//mybitread(tmp, variants->Subb, postmp, variants->subbJumpSize);
		tmp  = bitread ( variants->Subb, postmp, variants->subbJumpSize);
		jump = tmp;
	}
	
	postmp = absPos + jump; /** the starting offset in the bitarray of encoded variants */

	{	//Decoding until a '\0' (==> variant ==0) is found.
		uint variant;
		register uint ptr;
		THuff *huff = &(variants->H);
		len=0;
		ptr = postmp;	
		//ptr = decodeHuff(huff,&variant,variants->zoneMemHuff,ptr); 
			decodeNormalHuffMacro(huff,variant,(variants->zoneMemHuff),ptr); 
			len++;
		while (variant) {
			decodeNormalHuffMacro(huff,variant,variants->zoneMemHuff,ptr);
			//ptr = decodeHuff(huff,&variant,variants->zoneMemHuff,ptr);
			len++;	
		}		
	len--;
	}
	return (len);
}

/** asciiVariant 
	returns a pointer to a decoded variant (ended in '\0')
	*/
byte * asciiVariant(tVariants *variants, uint i, uint j) {
	byte *dst;
	register uint len=0;
	uint posSub, absPos;	
	uint jump;
	uint postmp;
	uint returnValue;
		
	postmp = i * variants->dirbElemSize;
	posSub  = bitread ( variants->Dirb, postmp, variants->dirbElemSize);
	//	mybitread(absPos,variants->Subb,posSub,variants->subbAbsSize);
	absPos = bitread (variants->Subb,posSub,variants->subbAbsSize);
	jump=0;
	if (j!=0) {
		uint tmp;
		postmp =  (posSub+ (variants->subbAbsSize) + (variants->subbJumpSize * (j-1)));
		//mybitread(tmp, variants->Subb, postmp, variants->subbJumpSize);
		tmp  = bitread ( variants->Subb, postmp, variants->subbJumpSize);
		jump = tmp;
	}
	postmp = absPos + jump; /** the starting offset in the bitarray of encoded variants */
	{	//Decoding until a '\0' (==> variant ==0) is found.
		uint variant;
		register uint ptr;
		THuff *huff = &(variants->H);
		dst = variants->decodeBuffer;
		ptr = postmp;	
			decodeNormalHuffMacro(huff,variant,variants->zoneMemHuff,ptr);
		//ptr = decodeHuff(huff,&variant,variants->zoneMemHuff,ptr);
		*dst=(byte)variant; dst++; len ++;
		while (variant) {
			decodeNormalHuffMacro(huff,variant,variants->zoneMemHuff,ptr);
			//ptr = decodeHuff(huff,&variant,variants->zoneMemHuff,ptr);
			*dst=(byte)variant; dst++;
			len++;
		}	
	len--;	
	}
	return ((byte*) (variants->decodeBuffer));
}


/** asciiVariant 
	returns a pointer to a decoded variant (ended in '\0')
	*/
byte * asciiVariantLen(tVariants *variants, uint i, uint j, uint *length) {
	byte *dst;
	register uint len=0;
	uint posSub, absPos;	
	uint jump;
	uint postmp;
	uint returnValue;
		
	postmp = i * variants->dirbElemSize;
	posSub  = bitread ( variants->Dirb, postmp, variants->dirbElemSize);
	//	mybitread(absPos,variants->Subb,posSub,variants->subbAbsSize);
	absPos = bitread (variants->Subb,posSub,variants->subbAbsSize);
	jump=0;
	if (j!=0) {
		uint tmp;
		postmp =  (posSub+ (variants->subbAbsSize) + (variants->subbJumpSize * (j-1)));
		//mybitread(tmp, variants->Subb, postmp, variants->subbJumpSize);
		tmp  = bitread ( variants->Subb, postmp, variants->subbJumpSize);
		jump = tmp;
	}
	postmp = absPos + jump; /** the starting offset in the bitarray of encoded variants */

	{	//Decoding until a '\0' (==> variant ==0) is found.
		uint variant;
		register uint ptr;
		THuff *huff = &(variants->H);
		dst = variants->decodeBuffer;
		ptr = postmp;	
		decodeNormalHuffMacro(huff,variant,variants->zoneMemHuff,ptr);
		//ptr = decodeHuff(huff,&variant,variants->zoneMemHuff,ptr);
		*dst=(byte)variant; dst++;	len ++;
		while (variant) {
			decodeNormalHuffMacro(huff,variant,variants->zoneMemHuff,ptr);
			//ptr = decodeHuff(huff,&variant,variants->zoneMemHuff,ptr);
			*dst=(byte)variant; dst++;
			len++;
		}	
	len--;	
	}

	*length=len;
	return ((byte*) (variants->decodeBuffer));
}



/** asciiVariantCopyBytes
	decodes a variant and copies the first "number" bytes to decodedVariant;
	*/
void asciiVariantCopyBytes(byte *decodedVariant, tVariants *variants, uint i, uint j, register uint number) {
	byte *dst;
	uint posSub, absPos;	
	uint jump;
	uint postmp;
	uint returnValue;
		
	dst =decodedVariant;
	postmp = i * variants->dirbElemSize;
	posSub  = bitread ( variants->Dirb, postmp, variants->dirbElemSize);
	//	mybitread(absPos,variants->Subb,posSub,variants->subbAbsSize);
	absPos = bitread (variants->Subb,posSub,variants->subbAbsSize);
	jump=0;
	if (j!=0) {
		uint tmp;
		postmp =  (posSub+ (variants->subbAbsSize) + (variants->subbJumpSize * (j-1)));
		//mybitread(tmp, variants->Subb, postmp, variants->subbJumpSize);
		tmp  = bitread ( variants->Subb, postmp, variants->subbJumpSize);
		jump = tmp;
	}
	postmp = absPos + jump; /** the starting offset in the bitarray of encoded variants */

	{	//Decoding until a '\0' (==> variant ==0) is found.
		uint variant;
		register uint ptr;
		register uint num;
		THuff *huff = &(variants->H);
		ptr = postmp;	

		for (num=0;num<number;num++) {
			decodeNormalHuffMacro(huff,variant,variants->zoneMemHuff,ptr);
			//ptr = decodeHuff(huff,&variant,variants->zoneMemHuff,ptr);
			*dst=(byte)variant; dst++;
		}	
	}	
}


/** Tells if the canonical word has either a unique variant or not **/
int hasUniqueVariant (tVariants *variants, uint idCanonical) {
		register uint dirbElemSize, offseti, offsetj, pos;
		dirbElemSize = variants->dirbElemSize;
		pos = dirbElemSize * idCanonical;
		uint *dirb = variants->Dirb;
		
		{
			offseti= bitread(dirb, pos, dirbElemSize);
			//mybitread(offseti, dirb, pos, dirbElemSize);
		}		
		pos+=dirbElemSize;  //idCanonical +1
		{
			offsetj= bitread(dirb, pos, dirbElemSize);
			//mybitread(offsetj, dirb, pos, dirbElemSize);
		}
		return ((offsetj - offseti) == (variants->subbAbsSize));
}




/** DEBUGING ONLY****************/

void showVariants2(tVariants *variants){
	uint i;
// 	fprintf(stderr,"\n zoneMem:\n");
// 	for (i=0;i<120; i++) {
// 		fprintf(stderr,"%c",variants->zoneMem[i]);
// 	}
// 	fprintf(stderr,"\n");
// 	for (i=0;i<120; i++) {
// 		if (!(i%10)) fprintf(stderr,"1");
// 		else fprintf(stderr," ");
// 	}

	fprintf(stderr,"\n Dir:\n");
	for (i=0;i<100; i++) {
		//fprintf(stderr,"[%d]",getKBitArray(variants->Dirb,i));
		uint posSub,postmp;
		postmp = i * variants->dirbElemSize;
		posSub  = bitread ( variants->Dirb, postmp, variants->dirbElemSize);
		fprintf(stderr,"[%d]", posSub);
	
	}
	fprintf(stderr,"\n SubDir: main be wrong if more than 1 variants in this part\n");
	for (i=0;i<10; i++) {
		uint posSub,postmp;
		postmp = i * variants->subbAbsSize;
		posSub  = bitread ( variants->Subb, postmp, variants->subbAbsSize);
		fprintf(stderr,"[%d]", posSub);
	}
	fprintf(stderr,"\n Decoded:\n");
	for (i=0;i<10; i++) {
		byte *dst;
		dst = asciiVariant(variants, i, 0);
		fprintf(stderr,"[String:%s]", dst);
		fprintf(stderr,"[Len:%d]", lenVariant(variants, i, 0)) ;
	}
	fflush(stderr);
}

// void showVariants(tVariants *variants){
// 	uint i;
// 	fprintf(stderr,"\n zoneMem:\n");
// 	for (i=0;i<120; i++) {
// 		fprintf(stderr,"%c",variants->zoneMem[i]);
// 	}
// 	fprintf(stderr,"\n");
// 	for (i=0;i<120; i++) {
// 		if (!(i%10)) fprintf(stderr,"1");
// 		else fprintf(stderr," ");
// 	}
// 	fprintf(stderr,"\n Dir:\n");
// 	for (i=0;i<1000; i++) {
// 		fprintf(stderr,"[%d]",variants->Dir[i]);
// 	}
// 	fprintf(stderr,"\n SubDir:\n");
// 	for (i=0;i<10; i++) {
// 		fprintf(stderr,"[%d]",variants->Sub[i]);
// 	}
// 	fflush(stderr);
// }

