#include "buildFacade.h"
#include "utils/ii.c"

/** Building the index */

    /* Creates index from text[0..length-1]. Note that the index is an 
      opaque data type. Any build option must be passed in string 
      build_options, whose syntax depends on the index. The index must 
      always work with some default parameters if build_options is NULL. 
      The returned index is ready to be queried. */

int build_index (uchar *text, ulong length, char *build_options, void **index) {
	int returnvalue;

    fprintf(stderr,"\n parameters (stopwordsFilename): \"%s\"\n",build_options); fflush(stderr);        
    //returnvalue = build_WCSA (text, length, build_options, index);
    return returnvalue;    
}

int build_index_baseFile(uchar *inbasename, uint blockSize, uint sampleK, char *stopwordsfile, void **index) {
	int returnvalue;
	
	fprintf(stderr,"\n *basename = %s, blocksize = %d, sampleK = %d\n",inbasename,blockSize,sampleK);
   returnvalue = build_WCSA (inbasename, blockSize,sampleK,stopwordsfile, index);
   return returnvalue;    
}


/**  Saves index on disk by using single or multiple files, having 
	proper extensions. */
int save_index (void *index, char *filename) {
	char *basename = filename;
	twcsa *wcsa=(twcsa *) index;

	uint i,j;
	char *outfilename;
	int file;
	char c;

	fprintf(stderr,"\nSaving structures to disk: %s.*\n",basename);	
			
	saveStopWords(wcsa->stopwordsfilename, basename);
	//the compressed text from ETDC
	saveTextInMem (wcsa->text, wcsa->textSize, basename);
	//the vocabulary from ETDC	
	saveVocabularyETDC(index, basename);	
	/** 2.b Saves the bitmap that indicates if a word (etdc voc) is indexable or not*/
	saveIndexedWordsVector(index, basename);		
	/** 2.c Saves the canonical words (ascii) into disk */	
	saveCanonicalsAscii(index, basename);
	/** 2.d Saves the ids of the variants of the canonical words into disk */
	saveVariantsCanonicals(index, basename);
	
	saveFreqVector    (wcsa->freqs,   wcsa->ncanonicals, basename);	
	// saves lenList Vector do disk
	savelenListVector (wcsa->lenList, wcsa->ncanonicals, basename);	
	// Saving the lists of occurrences 
	saveOccLists (index, basename);

	{uint totalSizeDisk=0;
		char * file;
		ulong memSize;
		file = (char *)malloc(sizeof(char)*(strlen(basename)+25));

		strcpy(file, basename); strcat(file, ".");	strcat(file, STOPWORDS_FILE_EXT);	
		totalSizeDisk += fileSize(file);		
		strcpy(file, basename); strcat(file, ".");	strcat(file, ETDCDATA);	
		totalSizeDisk += fileSize(file);
		strcpy(file, basename); strcat(file, ".");	strcat(file, ETDCVOC);				
		totalSizeDisk += fileSize(file);
		strcpy(file, basename); strcat(file, ".");	strcat(file, VOC_MARKED_INDEXED_WORDS);
		totalSizeDisk += fileSize(file);
		strcpy(file, basename); strcat(file, ".");	strcat(file, CANONICALS_EXT);
		totalSizeDisk += fileSize(file);
		strcpy(file, basename); strcat(file, ".");	strcat(file, VARIANTS_EXT);							
		totalSizeDisk += fileSize(file);
		strcpy(file, basename); strcat(file, ".");	strcat(file, FREQVECTOR);	
		totalSizeDisk += fileSize(file);		
		strcpy(file, basename); strcat(file, ".");	strcat(file, LENLISTS);	
		totalSizeDisk += fileSize(file);
		strcpy(file, basename); strcat(file, ".");	strcat(file, LISTSOCC);	
		totalSizeDisk += fileSize(file);		
			
		index_size(index, &memSize);
		fprintf(stderr,"\n Size in disk = %d bytes, sizeInMem = %d bytes", totalSizeDisk,memSize );
		free(file);
	}
	return 0;
}



    /**  Loads index from one or more file(s) named filename, possibly 
      adding the proper extensions. */
int load_index(char *filename, void **index){
	twcsa *wcsa;
	wcsa = loadWCSA (filename);
	(*index) = (void *) wcsa;	
	return 0;
}

	/** Frees the memory occupied by index. */
int free_index(void *index){
	twcsa *wcsa=(twcsa *) index;
//	ulong size;
//	index_size(index,&size);
//	printf("\n[destroying index] ...Freed %d bytes... RAM", size);	

	free(wcsa->words);
	free(wcsa->wordsZone.zone);
	free(wcsa->indexedWords);
	free(wcsa->canonicals);
	free(wcsa->canonicalsZone.zone);	
	free(wcsa->varsCanonicals);
	free(wcsa->varsETDC);
	
	free(wcsa->freqs);
	free(wcsa->lenList);
	free(wcsa->occs.occs);
	free(wcsa->occs.bcs);
	free(wcsa->occs.zoneOccs);
	free(wcsa->text);
	stdestroy_hash(wcsa->stopwordshash);
	if (wcsa->stopwordsfilename) free (wcsa->stopwordsfilename);
		
	free(wcsa);

	return 0;
}




	/** Gives the memory occupied by index in bytes. */

int index_size(void *index, ulong *size) {
	twcsa *wcsa=(twcsa *)index;
	*size=0;
/**/	*size += sizeof(twcsa);
	
/**/	*size += wcsa->textSize ;
/**/	*size += stmemused(wcsa->stopwordshash);

/**/	*size += wcsa->nwords * sizeof(uint *) ; //words
/**/  *size += wcsa->wordsZone.size;  //the zone of bytes where the "words" are kept.
/**/  *size += (((wcsa->nwords)+W-1)/W) * sizeof(uint); //the bitmap indicating which words are indexed.

/**/	*size += wcsa->ncanonicals * sizeof(uint *) ; //canonicals
/**/  *size += wcsa->canonicalsZone.size;  //the zone of bytes where the "canonicals" are kept.

/**/	*size += wcsa->ncanonicals * sizeof(uint); //ptrs to the codes with the ids of its variants.
/**/	*size += wcsa->varsETDCSize;               //ids of the variants of the canonicals compressed with ETDC.

/**/	*size += wcsa->ncanonicals * sizeof(uint) ;   //freqs
/**/	*size += wcsa->ncanonicals * sizeof(uint) ;   //lenList vect

/**/	*size += wcsa->ncanonicals * sizeof(uint) ;   //occs.occs
/**/	*size += (wcsa->occs.zoneOccsSize+W-1)/W * sizeof(uint);
/**/	*size += wcsa->occs.bcsSize ;	
	//printf("\n SIZE OF INDEX = %d bytes ",*size);
	return 0;	
}
                                                                                                 
                                                                                                                        
                                                                             



/** Querying the index =============================================================*/
	
	/* Writes in numocc the number of occurrences of the substring 
	   pattern[0..length-1] found in the text indexed by index. */
int count (void *index, uchar *pattern, ulong length, ulong *numocc){
	uint integerPatterns[MAX_INTEGER_PATTERN_SIZE];
	uint integerPatternSize;	

	twcsa *wcsa=(twcsa *) index;
	parseTextIntoIntegers(wcsa, pattern, length, integerPatterns, &integerPatternSize);	
	//no words
	if (!integerPatternSize) {*numocc=0; return 0;} //not found
	
	//only just one word
	if (integerPatternSize ==1) {*numocc= wcsa->freqs[integerPatterns[0]]; return 0;}
	
	//a phrase;
	{	uint *occs;	
		getPositions (index, integerPatterns, integerPatternSize, &occs, (uint *) numocc);
		free(occs);
	}			
	return 0;
}

	/* Writes in numocc the number of occurrences of the substring 
	  pattern[0..length-1] in the text indexed by index. It also allocates
	  occ (which must be freed by the caller) and writes the locations of 
	  the numocc occurrences in occ, in arbitrary order.  */	  
int locate(void *index, uchar *pattern, ulong length, ulong **occ, ulong *numocc){
	uint integerPatterns[MAX_INTEGER_PATTERN_SIZE];
	uint integerPatternSize,occurrences;	
	twcsa *wcsa=(twcsa *) index;
	
	parseTextIntoIntegers(wcsa, pattern, length, integerPatterns, &integerPatternSize);	
	
	if (!integerPatternSize) {*numocc=0; return 0;} //not found 

	getPositions (index, integerPatterns, integerPatternSize, (uint **) occ, (uint *) numocc);
	//fprintf(stderr,"\n** Pattern: %s appears %d times",pattern, *numocc);
	return 0;
}

   /* Gives the length of the text indexed */
int get_length(void *index, ulong *length) {
	twcsa *wcsa=(twcsa *) index;
	*length = wcsa->sourceTextSize;
	return 0;
}



/** ***********************************************************************************
  * Accessing the indexed text  
  * ***********************************************************************************/


	/**  Allocates snippet (which must be freed by the caller) and writes 
	  the substring text[from..to] into it. Returns in snippet_length the 
	  length of the text snippet actually extracted (that could be less 
	  than to-from+1 if to is larger than the text size). 			   */
int extract (void *index, ulong from, ulong to, uchar **snippet, ulong *snippet_length) {    		
    twcsa *wcsa=(twcsa *) index;    

   return -1;
}

  /** Displays the text (snippet) surrounding any occurrence of the 
    substring pattern[0..length-1] within the text indexed by index. 
    The snippet must include numc characters before and after the 
    pattern occurrence, totalizing length+2*numc characters, or less if 
    the text boundaries are reached. Writes in numocc the number of 
    occurrences, and allocates the arrays snippet_text and 
    snippet_lengths (which must be freed by the caller). The first is a 
    character array of numocc*(length+2*numc) characters, with a new 
    snippet starting at every multiple of length+2*numc. The second 
    gives the real length of each of the numocc snippets. */

int display (void *index, uchar *pattern, ulong length, ulong numc, 
        ulong *numocc, uchar **snippet_text, ulong **snippet_lengths) {

	twcsa *wcsa=(twcsa *) index;
				
	return -1;
}

    /**  Obtains the length of the text indexed by index. */

int length (void *index, ulong *length) {
	return (get_length(index,length));
}



/** ***********************************************************************************
  * WORD-ORIENTED QUERY FUNCTIONS: LocateWord and DisplayWord
  * ***********************************************************************************/  
	/* Writes in numocc the number of occurrences of the substring 
	  pattern[0..length-1] in the text indexed by index. It also allocates
	  occ (which must be freed by the caller) and writes the locations of 
	  the numocc occurrences in occ, in arbitrary order. These occurrences
	  refer to the offsets in TOH where the caller could start a display
	  operation. So locateWord implies synchronization using B.
	  Moreover, positions occ[numocc.. 2*numocc-1] is set with the rank in SE of the
	  words whose codes begin in TOH in the positions in occ[0... numocc-1]
	  ** Parameter kbefore sets locateWord not to obtain the offset in TOH of the
	     searched word, but the offset in TOH of k-before words before.	  
	*/	  
	  
int locateWord(void *index, uchar *pattern, ulong length, ulong **occ, ulong *numocc, uint kbefore){
	uint integerPatterns[MAX_INTEGER_PATTERN_SIZE];
	uint integerPatternSize,occurrences,l,r;	
	twcsa *wcsa=(twcsa *) index;
	
	parseTextIntoIntegers(wcsa, pattern, length, integerPatterns, &integerPatternSize);	
	if (!integerPatternSize) {*numocc=0; return 0;} //not found 

	return -1;
}  


  /** Displays the text (snippet) surrounding any occurrence of the 
    substring pattern[0..length-1] within the text indexed by index. 
    The snippet must include numc characters before and after the 
    pattern occurrence, totalizing length+2*numc characters, or less if 
    the text boundaries are reached. Writes in numocc the number of 
    occurrences, and allocates the arrays snippet_text and 
    snippet_lengths (which must be freed by the caller). The first is a 
    character array of numocc*(length+2*numc) characters, with a new 
    snippet starting at every multiple of length+2*numc. The second 
    gives the real length of each of the numocc snippets. */

 int displayWords (void *index, uchar *pattern, ulong length, ulong numc, 
         ulong *numocc, uchar **snippet_text, ulong **snippet_lengths, uint kbefore) {

 	twcsa *wcsa=(twcsa *) index;
 
 
 	return -1; 	     	
}





/** ***********************************************************************************
	  CONSTRUCTION OF THE INDEX 
    ***********************************************************************************/

int build_WCSA (uchar *inbasename, uint blockSize, uint sampleK, char *stopwordsfile, void **index) {
	twcsa *wcsa;
	wcsa = (twcsa *) malloc (sizeof (twcsa) * 1);
	*index = wcsa;
	stt_hash stopwordshash; 

	// Inicializes the arrays used to detect if a char is valid or not.
	StartValid();
	// Inicializes the arrays used translated a char into lowercase.
	StartToLow();
	
	//initializes the hash table of stopwords
	printf("\nLoading stop-words list...");
	loadStopWords(&stopwordshash,stopwordsfile);	
	wcsa->stopwordshash = stopwordshash;
	wcsa->stopwordsfilename = (char *) malloc (sizeof(char) * strlen(stopwordsfile)+1);
	strcpy(wcsa->stopwordsfilename, stopwordsfile);	

	wcsa->blockSize=blockSize;
	
	/** 1 ** Loads the compressed text into memory. */
	loadTextInMem(&(wcsa->text), &(wcsa->textSize),(char *)inbasename);	
	fprintf(stderr,"Loaded Compressed Text: %d bytes\n", wcsa->textSize);
	
	StartRusageTime();
	/** 2 ** loads the vocabulary of ETDC. Sets wordsZone, and words vector. */
	loadVocabularyETDC(wcsa, (char *)inbasename);
	fprintf(stderr,"Loaded Vocabulary of ETDC of text: %d bytes\n", wcsa->wordsZone.size);
		/** 2.b ** Creates the array that marks which words are indexed (valid) or not */
			markIndexedWords(*index);		
		
	  /** 2.c ** Obtains the vocabulary of (alphabetically) sorted canonical words 
							 So that function parseWordToCanonicalID(word) can work */
		{
		Wt_hash canonicalsHash;
		/** 2.c.1 ** Creates a temporal hash table to keep canonicals during construction */
		generateTempHashOfCanonicals(&canonicalsHash, *index);
			
		/** 2.c.2 ** Creates the array of sorted canonical words 
		             (and: byte **canonicals, byte *zoneCanonicals and ncanonicals)*/								
		getSortedCanonicalsArray(canonicalsHash, *index);											
		wtdestroy_hash(canonicalsHash);
		
				//		saveCanonicalsAscii (*index, "salida");
				//		free(wcsa->canonicals);
				//		free(wcsa->canonicalsZone.zone);
				//		loadCanonicalsAscii (*index, "salida");
				//		{int i;
				//			for (i=0;i<wcsa->ncanonicals;i++) printf("\n [%5d]:: canonical : %s ",i, wcsa->canonicals[i]);
				//		}
		
				//			{
				//				int pos;
				//				int i;
				//				for (i=0;i<wcsa->nwords; i++) {
				//					pos = parseWordToCanonicalID(wcsa, wcsa->words[i] );
				//					if (isIndexedWord(wcsa,i)) {
				//					//if (pos != -1) {
				//						printf("\n %d, word is %s ===> canonical : %s ",i, wcsa->words[i], wcsa->canonicals[pos]);	
				//					}
				//				}
				//			}		
		}

		//fprintf(stderr,"Seccion 2.d\n");
		/** 2.d ** Creates the compressed structure to keep the variants of each canonical word */
		{
			register uint i;
			register uint ncanonicals = wcsa->ncanonicals;
			register uint nwords = wcsa->nwords;
		/** 2.d.1 ** tmpNumVariants[i] = number of variants of each canonical $i$ */			
			uint *tmpNumVariants = (uint *) malloc (sizeof(uint) * ncanonicals);
			{	
				int x;
				for (i=0;i<ncanonicals;i++) {tmpNumVariants[i]=0;}
				for (i=0;i<nwords;i++) {
					if (isIndexedWord(wcsa,i)) {
						x=parseWordToCanonicalID(wcsa,wcsa->words[i]);
						tmpNumVariants[x]++;
					}
				}
			}

		//fprintf(stderr,"Seccion 2.d.2\n");			
		/** 2.d.2 ** tmpPosVariants[i][j] == rank of the j^{th} variants of the i^{th} canonical word */				
			uint **tmpPosVariants; 
			{  
				int x;
				//tmp array to count how many variants have already been added to a given canonical word.
				uint *addedVariants = (uint *) malloc (sizeof(uint) * ncanonicals);

				//allocates memory and initializes addedVariants[i] = 0 \forall i
				tmpPosVariants = (uint **) malloc(sizeof(uint*) * ncanonicals);
				for (i=0;i<ncanonicals;i++) {
					tmpPosVariants[i] = (uint *) malloc (sizeof(uint)*tmpNumVariants[i]);
					addedVariants[i] =0;
				}

				//finds out the variants (its position in words array) for each canonical word
				for (i=0;i<nwords;i++) {
					if (isIndexedWord(wcsa,i)) {
						x=parseWordToCanonicalID(wcsa,wcsa->words[i]); //x would be "-1" if not indexed word.
						tmpPosVariants[x][addedVariants[x]++]= i;
					}					
				}

				free(addedVariants);
				
				//at this point:
				//  - tmpNumVariants[i] keeps the number of variants for the $i^{th}$ canonical word.
				//  - tmpPosVariants[i][j] keeps the rank in "words array" of the j^{th} variants of the $i^{th}$ canonical word.
				
														//				{int j;
														//					for (i=0;i<ncanonicals;i++) {
														//							printf("\n **Canonical= %s** %d variants ==>",wcsa->canonicals[i], tmpNumVariants[i]);
														//							for (j=0;j<tmpNumVariants[i]; j++) {
														//								printf("<<%s>>",wcsa->words[tmpPosVariants[i][j]]);
														//							}
														//					}					
														//				}
			}
		//fprintf(stderr,"Seccion 2.d.3\n");
		/** 2.d.3 ** tmpPosVariants[i][j] == rank of the j^{th} variants of the i^{th} canonical word 
			           tmpPosVariants are encoded (with ETDC); structures used are: varsETDC, variantCodes and barsETDCSize; */
			
			{uint i;
			 uint max =0;
			 	for (i=0;i<wcsa->ncanonicals;i++) {
			 		if (tmpNumVariants[i] > tmpNumVariants[max]) max=i;
				}
				max=tmpNumVariants[max];
				printf("\n @@@@@@@the max number of variants is %d\n\n",max);
			}
			
			encodePosOfVariantsOfCanonical(*index, tmpPosVariants, tmpNumVariants);
			for (i=0;i<ncanonicals;i++) {
					free(tmpPosVariants[i]);
				}
			free(tmpPosVariants);
			free(tmpNumVariants);			
			//			saveVariantsCanonicals (*index, (char *)inbasename);
			//			free(wcsa->varsCanonicals);
			//			free(wcsa->varsETDC);
			//			loadVariantsCanonicals (*index, (char *)inbasename);
										
		}		
		
		StopRusageTime("tiempos", "STEMMING");
		//fprintf(stderr,"Seccion 3\n");				
	{	/** 3 ** computes frequencies of each word. */
		int *mapToCanonical;	//maps a word from "words[i]" into canonicals[j].  -1 if "i" is not indexable
		wcsa->freqs = (uint *) malloc (sizeof(uint) * wcsa->ncanonicals);
	 	computeFreqs (*index, &mapToCanonical); //computes freq and initializes mapToCanonical  
	

		//fprintf(stderr,"Seccion 4\n");
		/** 4 ** creates a temporal list of occurrences of each word (block-oriented).
					gives also the len of each list */
		uint i;
		uint **occList;
		uint *lenList;
		createListsOfOccurrences (*index, &occList, &lenList, mapToCanonical);
		wcsa->lenList = lenList;
		free(mapToCanonical); 

			
		//fprintf(stderr,"Seccion 5\n");
		/** 5 ** Encoding the occurrences 
					temporal zoneOccsTmp is create [uint M | uint O][uint M | uint O]...
					bc is set.
					O's are definitive offsets into bc ==> sizeBcs is computed
					M's are definitive offsets into text.
					*/
		wcsa->occs.sampleK = sampleK;
		wcsa->occs.occs = (uint *) malloc (sizeof(uint) * wcsa->ncanonicals);

		//fprintf(stderr,"Seccion 5.a\n");
			/** 5.a ** computes the number of Absolute samples to keep. Depends on sampleK. */
			uint absCounter =0;
			{  uint ncanonicals = wcsa->ncanonicals;
				for (i=0;i<ncanonicals; i++) absCounter += ( (wcsa->lenList[i] + sampleK -1) /sampleK);	
			//	fprintf(stderr,"\n Number of absolute samples to store is = %d",absCounter);		 					
			}			
		uint *zoneOccsTmp = (uint *) malloc (sizeof(uint) * 2 * absCounter);  // [abs|offsetBC]

	//	fprintf(stderr,"Seccion 5.b\n");
			/** 5.b ** computes the size of bc vector by simulating encoding process of gaps */
			uint bcssize=0;
			{
				uint ncanonicals = wcsa->ncanonicals;
				uint id,j;
				uint prevvalue, gap, numbytes;
				uint len ;
	//			fprintf(stderr,"ncanonicals: %d\n",ncanonicals);
				uint counterBCcodes =0; uint counterAbsSamples=0;
				for (id=0; id<ncanonicals; id++) {
					len= wcsa->lenList[id];
		//			fprintf(stderr,"id: %d lesnList %d\n",id,len);
					wcsa->occs.occs[id]=counterAbsSamples;   /** points to a number of bucket --> changes in step **6** */
					counterAbsSamples++;
					prevvalue= occList[id][0];
			//						fprintf(stderr,"id: %d counterAbs %d\n",id,counterAbsSamples);
					for (j=1; j<len; j++){
						if (j %sampleK) { //a gap;
							gap = occList[id][j]-prevvalue;
							prevvalue = occList[id][j];
							SIZE_DEC_TO_BC(gap,numbytes);
							bcssize += numbytes;
							counterBCcodes++;
						}
						else {
							prevvalue = occList[id][j];
							counterAbsSamples++;
						}
					}
				}
		//		fprintf(stderr,"\n Number of absolute samples to store is = %d",counterAbsSamples);						
		//		fprintf(stderr,"\n Number of gaps to store is = %d",counterBCcodes);	
		//		fprintf(stderr,"\n Size of bcs Vector should by  %d bytes",bcssize);
				//fprintf(stderr,"\n [SHOULD BE IDENTICAL (not in block-oriented)]: number of words = %d, and sum of occurs is %d",wcsa->Nwords, counterAbsSamples + counterBCcodes);	
			}
		wcsa->occs.bcsSize = bcssize;
		wcsa->occs.bcs = (byte *) malloc (sizeof(byte) * wcsa->occs.bcsSize);									

//		fprintf(stderr,"Seccion 5.c\n");
		/** 5.c ** encodes gaps into bcs && set pointers into zoneoccurssTmp */
			{
				uint ncanonicals = wcsa->ncanonicals;
				byte *bcs = wcsa->occs.bcs;
				uint id,j;
				uint prevvalue, gap;
				uint len ;

				uint bcpos=0;
				uint zoneoccspos=0;
				
				for (id=0; id<ncanonicals; id++) {
					len= wcsa->lenList[id];						
					for (j=0; j<len; j++){
						if (j %sampleK) { //a gap;
							gap = occList[id][j]-prevvalue;
							prevvalue = occList[id][j];
							DEC_TO_BC(bcs, bcpos, gap);  //bcpos is increased.
						}
						else {
							prevvalue = occList[id][j];
							zoneOccsTmp[zoneoccspos++]=prevvalue;
							if (len ==(j+1)) zoneOccsTmp[zoneoccspos++] = FILLERUINT;  //fill next int (last occurr in abs sample).
							else zoneOccsTmp[zoneoccspos++]= bcpos;							
						}
					}
				}
	//			fprintf(stderr,"\n Gaps were encoded with bc ==> %d bytes",bcpos);
	//			fprintf(stderr,"\n zoneoccspos = %d ",zoneoccspos);				
			}

		wcsa->occs.absOffsetSize = bits(wcsa->textSize);    //in bits
		wcsa->occs.bcOffsetSize = bits(wcsa->occs.bcsSize); //in bits
		wcsa->occs.occsBucketSize = wcsa->occs.absOffsetSize + wcsa->occs.bcOffsetSize ; //in bits
		wcsa->occs.numBuckets   = absCounter;
		wcsa->occs.zoneOccsSize = wcsa->occs.numBuckets * wcsa->occs.occsBucketSize; //in bits
		wcsa->occs.zoneOccs = (uint *) malloc (sizeof(uint) * ((wcsa->occs.zoneOccsSize + W -1)/W) );
		wcsa->occs.zoneOccs[((wcsa->occs.zoneOccsSize +W-1)/W)-1] =0000; //so valgrind is shut up		

		//fprintf(stderr,"\n #### zoneOccSize= %d bits. allocated %d bytes para zoneOccs ",wcsa->occs.zoneOccsSize, sizeof(uint) * ((wcsa->occs.zoneOccsSize + W -1)/W) );
		//exit(0);

		fprintf(stderr,"Seccion 6\n");
		/** 6 ** encode zoneoccurssTmp into zoneoccurss using a "kbit-fashion" 
		         sets also offsets in wcsa->occs.occs[id] \forall id;            */
			{	uint i,value;
				uint ncanonicals = wcsa->ncanonicals;
				uint numBuckets = wcsa->occs.numBuckets;
				uint absBits = wcsa->occs.absOffsetSize;
				uint bcBits = wcsa->occs.bcOffsetSize;
				uint pos=0;
		//		fprintf(stderr,"\n setting offsets in wcsa->occs.occs vector");
				for (i=0; i<ncanonicals; i++) wcsa->occs.occs[i]*= (wcsa->occs.occsBucketSize);
		//		fprintf(stderr,"\n enconding directory of [Abs|bc] buckets to bit-oriented form");
				for (i=0; i<numBuckets; i++) {
					value = zoneOccsTmp[i*2];
					bitwrite (wcsa->occs.zoneOccs, pos, absBits, value);	pos += absBits;
					value = zoneOccsTmp[i*2+1];
					bitwrite (wcsa->occs.zoneOccs, pos, bcBits, value);	pos += bcBits;
				}				
			}

		/** checking that the lists are decoded successfully *************/

//		fprintf(stderr,"\n ! CHECKING that the decoded lists are identical to the original ones. ");
		{ //checking:: decoding a list
			uint i, id ,len;
			uint *list;
			
			for (id=0;id < wcsa->ncanonicals;id++) {
				uncompressList (*index, id, &list, &len);			
				for (i=0;i<len;i++){
					if (list[i] != occList[id][i]) {
	//					fprintf(stderr,"\n decoding of lists failed for id = %d: DIFFERENT!!",id);
						exit(0);
					}
				} 
				
				free(list);				
			}			
		}
	//	fprintf(stderr,"\n ! Decoding the list of occurrences worked fine (the same as previous values) ");
		
//CHECKING : DEC:DODE ALL THE BLOCK-POSITIONS IN THE INDEX AND PERFOR A HORSPOOL ALGOR OVER THE POINTED BLOCKS.
//		{  //checking it worked: chooses an id. And decodes from the pointed offsets on.
//			//all the decoded words should be the same.
//			uint absBits = wcsa->occs.absOffsetSize;
//			uint bcBits = wcsa->occs.bcOffsetSize;
//			byte *bcs = wcsa->occs.bcs;
//			byte *text = wcsa->text;
//		
//			uint id =24999;
//			uint i, pos, num;
//			uint freq = wcsa->freqs[id];
//			uint *list; uint len;
//			uncompressList (*index, id, &list, &len);		
//	
//			fprintf(stderr,"\n ! Decoding all the %d occurrences of word[%d]= <<%s>>",freq, id, wcsa->words[id]);fflush(stderr);		 		
//		
//			for (i=0; i<freq;i++) {
//				pos =list[i];									
//				fprintf(stderr,"\n   [%d] decoded word at position text[%d] ",i,pos);
//				pos = decode (text, pos, &num) ;	
//				fprintf(stderr," and id = %d is ",num);fflush(stderr);
//				fprintf(stderr,"<<%s>>",wcsa->words[num]);fflush(stderr);
//	 			
//			}			
//			free(list);
//		}			

		//check_BC_works(1000000);
		/******************************************************/

//		for (i=0;i<wcsa->ncanonicals;i++) {
//				printf("\n freq[%d] = %d",i,wcsa->freqs[i]);
//		}

		//frees memory
		for (i=0;i<wcsa->ncanonicals;i++) free(occList[i]);
		free(occList);
		free(zoneOccsTmp);
		

	}

	fprintf(stderr,"\n The index has already been built!!\n");
	

	//DESCOMPRIMIR TODAS AS LISTAS E COMPARALAS CO UINT ** occList  (free en linea 387))


	return 0;
}




/** ********************************************************************
  * Loading from disk
  **********************************************************************/ 

/**-----------------------------------------------------------------  
 * LoadWCSA.                                                       
 * Loads all the data structures of WCSA (included the icsa)     
 ----------------------------------------------------------------- */ 

twcsa *loadWCSA(char *filename) {
	twcsa *wcsa;
	wcsa = (twcsa *) malloc (sizeof (twcsa) * 1);
	void *Index = (void *) wcsa;
	
	// Inicializes the arrays used to detect if a char is valid or not.
	StartValid();
	// Inicializes the arrays used translated a char into lowercase.
	StartToLow();

	
	/** 1 ** Loads the compressed text into memory. */
	loadTextInMem(&(wcsa->text), &(wcsa->textSize),(char *)filename);	
	fprintf(stderr,"Loaded Compressed Text: %d bytes\n", wcsa->textSize);
	
	/** 1.b Loads the list of stopwords into a hash table. */
	{ char stopfile[1000];
		strcpy(stopfile, filename);	strcat(stopfile, "."); strcat(stopfile, STOPWORDS_FILE_EXT);
		loadStopWords(&(wcsa->stopwordshash),stopfile);
		wcsa->stopwordsfilename = NULL; //not needed during searches !! (only for saves_index)
	}		
		
	/** 2 ** loads the vocabulary of ETDC. Sets HT, zonewords, words vector, nwords */
	loadVocabularyETDC (Index, filename);
	fprintf(stderr,"Loaded Vocabulary of ETDC of text: %d bytes\n", wcsa->wordsZone.size);

	/** 2.b Loads the bitmap that indicates if a word (etdc voc) is indexable or not*/
	loadIndexedWordsVector(Index, filename);
		
	/** 2.c Loads the canoncal words (ascii) into wcsa. */
	loadCanonicalsAscii (Index, filename);

	/** 2.d Saves the ids of the variants of the canonical words into disk */
	loadVariantsCanonicals(Index, filename);
	
	/** 3 ** Loading freq vector */
	{uint size; //the size of the vocabulary in #ofwords =>> already init in "loadvocabulary"
	loadFreqVector(&(wcsa->freqs), &size, (char *)filename);	
	}		

	/** 4 ** Loading lenList vector */
	{uint size; //the size of the vocabulary in #ofwords =>> already init in "loadvocabulary"
	loadlenListVector(&(wcsa->lenList), &size, (char *)filename);	
	}		

	/** 5 ** Loading the lists of occurrences */
	loadOccLists (Index, filename);					    
	return wcsa;
}


/** ****************************************************************
  * Querying the index WCSA
  * ***************************************************************/
///////////////////////////////////////////////////////////////////////////////////////
//					   FUNCTIONS NEEDED FOR SEARCHING A PATTERN    					 //
///////////////////////////////////////////////////////////////////////////////////////

void printWord(byte *aWord, uint len) {
		uint i;
		for (i=0;i<len;i++) printf("%c",aWord[i]);
}

/*------------------------------------------------------------------
 * Given a text pattern translates it into a list of integers (corresponding to the
 * canonical words associated to the valid words in the text pattern) 
 ------------------------------------------------------------------*/
void parseTextIntoIntegers(twcsa *wcsa, byte *textPattern, uint patLen, uint *integerPattern, uint *sizeIntegers) {
	
	byte *pbeg,*pend,*wordstart,*aWord;
	register unsigned long size;
	tCanonical canonical;	
	uint index =0; 
		 	
	pbeg = textPattern; 
	pend = pbeg + patLen;
	
	//printf("\n### TO PARSE :");printWord(textPattern,patLen);
							
	while (pbeg <pend) {  
		//parsing a word or separator
		size=0;
		wordstart = pbeg;
		if (_Valid[*pbeg]) {   //alphanumerical data
			while ( (size<MAX_SIZE_OF_WORD) && (pbeg<=pend) && ( _Valid[*pbeg] )) {size++;pbeg++;}
		 }
		else {  // a separator
			while ( (size<MAX_SIZE_OF_WORD) && (pbeg<=pend)&& (!_Valid[*pbeg]) ) {size++;pbeg++;}
		}
		
		if ( pbeg >= pend ) {  //needed in case of the last word been parsed further away of cend. 
			size -= (pbeg-pend);		  			
			pbeg=pend;
		}					
		  				
		
		//The parsed word is "aWord", and its length is "size"...
		aWord=wordstart;																		
		//printf("\n##### PARSED WORD:<<");printWord(aWord,size);printf(">> of size %d",size);
		if (_Valid[*aWord]) { //either a word or a stopword		
			register uint min,max,p;	
			byte *canonicWord;		
			getCanonical(&canonical, aWord,size, wcsa->stopwordshash);
			canonicWord = canonical.word;					
			
			if (canonical.type != IS_STOPWORD)  { // a "meaningfull word" (a non-stopword)
				// Binary search on the canonical words (wordsData)
				min = 0;
				max = (wcsa->ncanonicals) - 1;
				while(min < max) {
					p = (min+max)/2;
					//printf("\n\n last pivot = %d and word there is <<%s>>\n",p,(char*)wcsa->canonicals[p]);
					if(strcmp((char*)canonicWord, (char*)wcsa->canonicals[p]) > 0) min = p+1;
					else max = p;
				}
				
				//printf("\n\n last pivot = %d and word there is %s\n",min,(char*)wcsa->canonicals[min]);
				if(!strcmp((char*)canonicWord, (char*)wcsa->canonicals[min])){
					 integerPattern[index++] = min; // min +1;  /* OJO !! CAMBIADO 12-MAY-2008 */
					//printf(" %s :its id = %d",wcsa->canonicals[min],min);
				}
				else {
					//printf("@@@@"); printWord(canonicWord,size);printf("<<<"); printf("%s",wcsa->canonicals[min]); printf(">>> does not appear", aWord);
					*sizeIntegers = 0; return;
				}  // a valid word that does not appear in the source text.
				
			}								
		}		
	}// end while
	*sizeIntegers = index;		
	
	
	//	{uint i;
	//		printf(" THE OBTAINED IDS ARE ==> ");
	//		if (index) for (i=0;i<index;i++) printf("[%d]",integerPattern[i]);
	//			printf("\n");
	//	}

}

// This is now in file "filterWord"
//
///** returns the ID (into the canonical array) of a variant */
////int parseWordToCanonicalID(twcsa *wcsa, byte *aWord, uint size) {
//int parseWordToCanonicalID(twcsa *wcsa, byte *aWord) {
//	tCanonical canonical;	
//		 						  																			
//		if (_Valid[*aWord]) { //either a word or a stopword		
//			register uint min,max,p;	
//			byte *canonicWord;		
//			//getCanonical(&canonical, aWord,size, wcsa->stopwordshash);
//			getCanonicalNoSize(&canonical, aWord, wcsa->stopwordshash);
//			canonicWord = canonical.word;					
//			
//			if (canonical.type != IS_STOPWORD)  { // a "meaningfull word" (a non-stopword)
//				// Binary search on the canonical words (wordsData)
//				min = 0;
//				max = (wcsa->ncanonicals) - 1;
//				while(min < max) {
//					p = (min+max)/2;
//					if(strcmp((char*)canonicWord, (char*)wcsa->canonicals[p]) > 0) min = p+1;
//					else max = p;
//				}
//				if(!strcmp((char*)canonicWord, (char*)wcsa->canonicals[min])) return min; //( min + 1 );
//				else {return -1;}  // a valid word that does not appear in the source text.
//			}			
//			else return -1;				
//		}
//		else return -1;
//}





	/** ------------------------------------------------------------------
	 * Returns the number of occurrences of a given text pattern
	 *------------------------------------------------------------------ */
int countTextOcurrences(twcsa *wcsa, byte *textPattern) {
	return -1;
}


	/** ------------------------------------------------------------------
	 * locateTextOcurrences: 
	 * Returns the offsets of the source text where a word/phrase appears
	 * Returns also the number of occurrences.
	 *------------------------------------------------------------------ */
uint *locateTextOcurrences(twcsa *wcsa, byte *textPattern, int *numberOccurrences) {

}



	/** ------------------------------------------------------------------
	 * displayTextOcurrences:
	 * Shows in stdout, the text around the occurrences of a word/phrase
	 * Returns also the number of occurrences.
	 * radixDisplay is the numberOfWords from that are decoded previosly and after each occurrence
	 *------------------------------------------------------------------ */
	 
int  displayTextOcurrencesNoShow(void *index, uchar *pattern, ulong length, uint radixDisplay, uint maxnumc) {

	uint integerPatterns[MAX_INTEGER_PATTERN_SIZE];
	uint integerPatternSize,occurrences;	

	twcsa *wcsa=(twcsa *) index;
	 	byte *text_aux;
	uchar c;

	parseTextIntoIntegers(wcsa, pattern, length, integerPatterns, &integerPatternSize);	
	if (!integerPatternSize) {return 0;} //not found 

	uint *list, numocc, pos, i ,z, num;
	uint previousAlfanumerical;
	byte *buff;
	buff = wcsa->text;	
	
	getPositions (index, integerPatterns, integerPatternSize, &list, &numocc);
	
	//fflush(stdout);fflush(stderr);
	printf(" %6d, ** Pattern: <<%s>>  ", numocc,pattern);
//	printf("\n ** Pattern: <<%s>>  appears %d times: in positions",pattern, numocc);
//	for (i=0;i<numocc;i++) printf("<%d>",list[i]);
//	printf("\n");

	ulong maxsnippetLen  = maxnumc;
	ulong extractedbytes = 0;
	uint snippetLen;
	text_aux = (byte *) malloc (maxsnippetLen+1);

	for (i=0;i<numocc;i++) {
		pos =list[i];									
		
		
		
		//moves "radixDisplay"  words before.
				//		z = radixDisplay;		
				//		if (pos >0) pos--; //previous stopper
				//		while ((z>0) && (pos >0)) {			
				//			if (buff[pos] <128) { // a stopper
				//				pos--; 
				//			}
				//			//a continuer ?
				//			while((pos >0)&&(buff[pos] >=128))pos--;			
				//			z--;
				//		} 
				//		if (z==0) pos++;
					
		//moves backwards "radixWords VALID words "
		{	
			int iters = radixDisplay;
			int x;
			uint returnedPos = pos;

				
			if (iters > 0) {		
				x=pos-2; //byte prior to prev stopper
				while ((x>=0) && (buff[x]>=128)) x--;  //points to prev stopper.
				
				register uint j=1;
				uint aID;   // an offset in words array
				int p=x+1;
						
				while ((j<=iters) && (p>=0)){
					p=decode(buff,p,&aID);
					if (isIndexedWord(wcsa,aID)) {
						j++;  //continues matching process but 1 word less left.		
						returnedPos = x+1;	
					}
					
					//moves to the beginning of the prev code.
					x--;
					while ((x>=0) && (buff[x] >=128)) x--;
					p=x+1;			
				}//while
				
				pos=returnedPos;	
			}	
		}	
		//Already synchronized "numWords before"
	

		snippetLen=0;
		byte *dst =text_aux;
		byte *src;
		
		//decoding the 2*radixWords [valid/any kind of] words.
		previousAlfanumerical=0;																			
		z=0;                                                      		
		uint end = wcsa->textSize;                                		
				                                                      		
		while ((pos < end) && (z<radixDisplay*2)) {               		
			pos = decode (wcsa->text, pos, &num) ;    
			if (isIndexedWord(wcsa,num)) {      
				z++;                              
			}                                   
			              		
			//writes space or not                                   		
			if (_Valid[wcsa->words[num][0]] ) {                     		
				if (previousAlfanumerical) {                          		
					//printf(" ");                                        		
					//c= ' '; 
					*dst = ' '; dst++;  snippetLen++;                                          		
				}                                                     		
				previousAlfanumerical = 1;                            		
			}                                                       		
			else {                                                  		
				previousAlfanumerical = 0;                        
			}  
			                                                 
			//printf("%s",wcsa->words[num]);                      
			{uchar *buf = wcsa->words[num];                     
			 while (*dst++ = *buf++) snippetLen++;                                
			}                                                   
			//z++;                                                
		}                                                     
	    
	  extractedbytes += snippetLen;           
	  /**DOES NOT CHECK "END OF BUFFER" !!!!!!!!!!!!!!!!!!! */                                         
	} //for                                                      
	                                                        
	if (numocc) free(list);
	
	if (text_aux) free (text_aux);
	return extractedbytes; 	
}
	 
//int  displayTextOcurrencesNoShow(void *index, uchar *pattern, ulong length, uint radixDisplay) {
//
//	uint integerPatterns[MAX_INTEGER_PATTERN_SIZE];
//	uint integerPatternSize,occurrences;	
//
//	twcsa *wcsa=(twcsa *) index;
//	uchar c;
//
//	parseTextIntoIntegers(wcsa, pattern, length, integerPatterns, &integerPatternSize);	
//	if (!integerPatternSize) {return 0;} //not found 
//
//	uint *list, numocc, pos, i ,z, num;
//	uint previousAlfanumerical;
//	byte *buff;
//	buff = wcsa->text;	
//	
//	getPositions (index, integerPatterns, integerPatternSize, &list, &numocc);
//	
//	//fflush(stdout);fflush(stderr);
//	printf(" %6d, ** Pattern: <<%s>>  ", numocc,pattern);
////	printf("\n ** Pattern: <<%s>>  appears %d times: in positions",pattern, numocc);
////	for (i=0;i<numocc;i++) printf("<%d>",list[i]);
////	printf("\n");
//	for (i=0;i<numocc;i++) {
////		printf("\n[%3d]",i);
//		pos =list[i];									
////		printf(", pos:[%d]\n\t<<",pos);
//		
//		//move radixDisplay "valid" ords before.
//		z = radixDisplay;		
//		if (pos >0) pos--; //previous stopper
//		while ((z>0) && (pos >0)) {			
//			if (buff[pos] <128) { // a stopper
//				pos--; 
//			}
//			//a continuer ?
//			while((pos >0)&&(buff[pos] >=128))pos--;			
//			z--;
//		} 
//		if (z==0) pos++;
//		
//		//decoding
//		previousAlfanumerical=0;
//		z=0;
//		uint end = wcsa->textSize;
//		
//		
//		while ((pos < end) && (z<radixDisplay*2)) {
//			pos = decode (wcsa->text, pos, &num) ;
//			//writes space or not
//			if (_Valid[wcsa->words[num][0]] ) {
//				if (previousAlfanumerical) {
////					printf(" ");
//					c= ' ';
//				}
//				previousAlfanumerical = 1;
//			}
//			else {
//				previousAlfanumerical = 0;
//			}
////			printf("%s",wcsa->words[num]);
//			{uchar *buf = wcsa->words[num];
//			 while (c= *buf++) ;
//			}
//			z++;
//		}
////		printf(">>");
//	
//	}
//	
//	if (numocc) free(list);
//	
//	return numocc;
//	
//}

int displayTextOcurrences(void *index, uchar *pattern, ulong length, uint radixDisplay) {

	uint integerPatterns[MAX_INTEGER_PATTERN_SIZE];
	uint integerPatternSize,occurrences;	

	twcsa *wcsa=(twcsa *) index;

	parseTextIntoIntegers(wcsa, pattern, length, integerPatterns, &integerPatternSize);	
	if (!integerPatternSize) {return 0;} //not found 

	uint *list, numocc, pos, i ,z, num;
	uint previousAlfanumerical;
	byte *buff;
	buff = wcsa->text;	
	
	getPositions (index, integerPatterns, integerPatternSize, &list, &numocc);
	
	fflush(stdout);fflush(stderr);
	printf("\n** Pattern: %s appears %d times: in positions",pattern, numocc);
	for (i=0;i<numocc;i++) printf("<%d>",list[i]);
	printf("\n");
	
	for (i=0;i<numocc;i++) {
		printf("\n[%3d]",i);
		pos =list[i];									
		printf(", pos:[%d]\n\t<<",pos);
		
		
			
			//move "radixDisplay"  words before.
					//		z = radixDisplay;		
					//		if (pos >0) pos--; //previous stopper
					//		while ((z>0) && (pos >0)) {			
					//			if (buff[pos] <128) { // a stopper
					//				pos--; 
					//			}
					//			//a continuer ?
					//			while((pos >0)&&(buff[pos] >=128))pos--;			
					//			z--;
					//		} 
					//		if (z==0) pos++;
						
			//moves backwards "radixWords VALID words "
		{	
			int iters = radixDisplay;
			int x;
			uint returnedPos = pos;
				
			if (iters > 0) {		
				x=pos-2; //byte prior to prev stopper
				while ((x>=0) && (buff[x]>=128)) x--;  //points to prev stopper.
				
				register uint j=1;
				uint aID;   // an offset in words array
				int p=x+1;
						
				while ((j<=iters) && (p>=0)){
					p=decode(buff,p,&aID);
					if (isIndexedWord(wcsa,aID)) {
						j++;  //continues matching process but 1 word less left.		
						returnedPos = x+1;	
					}
					
					//moves to the beginning of the prev code.
					x--;
					while ((x>=0) && (buff[x] >=128)) x--;
					p=x+1;			
				}//while
				
				pos=returnedPos;	
			}	
		}	
	
		
		//decoding the 2*radixWords [valid/any kind of] words.
		previousAlfanumerical=0;
		z=0;
		uint end = wcsa->textSize;
				
		while ((pos < end) && (z<radixDisplay*2)) {
			pos = decode (wcsa->text, pos, &num) ;
			if (isIndexedWord(wcsa,num)) {
				z++;
			}
			//writes space or not
			if (_Valid[wcsa->words[num][0]] ) {
				if (previousAlfanumerical) {
					printf(" ");
				}
				previousAlfanumerical = 1;
			}
			else {
				previousAlfanumerical = 0;
			}
			printf("%s",wcsa->words[num]);
			//z++;
		}
		printf(">>");
	
	}
	
	if (numocc) free(list);
	
	return numocc;
	
}


/** FUNCTIONS NEEDED DURING SEARCHES */




	/** ------------------------------------------------------------------
	 * recovers the source text by calling display(0,fileSize);
	 * ------------------------------------------------------------------ */
void recoverSourceText1(twcsa *wcsa, char *basename, char *ext, uint sourceTextSize) {		
	int start;int end;
	byte *cc;
	char *filename = (char *) malloc (strlen( basename)+ strlen(ext)+1);
	ulong length;
	
	strcpy( filename,  basename);
	strcat( filename, ext);
	filename[strlen( basename)+ strlen(ext)]='\0';
	fprintf(stderr,"\n uncompressing text into file -->%s ",filename);fflush(stderr);

	FILE *salida;
	unlink( filename);
	salida = fopen( filename,"w");
	start=0; end = sourceTextSize-1;
	

	cc = (byte *) malloc (sourceTextSize* sizeof(uchar));
	ulong leng;
//	displayFacade(wcsa,start,end,&leng,cc);
	fprintf(stderr,"\n sourceTextSize = %d, len = %d",sourceTextSize,leng);

	fwrite(cc,sizeof(byte),leng,salida);
	fclose(salida);

	free(cc);
	free(filename);
}





double getTime2 (void)
{
	double usertime, systime;
	struct rusage usage;

	getrusage (RUSAGE_SELF, &usage);

	usertime = (double) usage.ru_utime.tv_sec +
		(double) usage.ru_utime.tv_usec / 1000000.0;
	systime = (double) usage.ru_stime.tv_sec +
		(double) usage.ru_stime.tv_usec / 1000000.0;

	return (usertime );//+ systime);
}



/**------------------------------------------------------------------ 
  *  MAIN PROGRAM: FOR BUILDING THE INDEX
  *------------------------------------------------------------------ */
#ifdef FACADEWITHMAIN
	int main(int argc, char* argv[])
	{
		
		char *infile, *outbasename;	// Name of in/out files
		char *stopwordsfile;
		byte *inputBuffer;
		ulong finsize;
	
	 	int f_in;
		void *Index;

		uint ksample, blocksize;
		
		printf("\n*Presentation level for CSA");
		printf("\n*CopyRight (c) 2007 [LBD & G.N.]\n\n");
	
		// Reads input parameters from command line.
		if(argc < 6) {
			printf("Use: %s <in file> <out basename> <kSample> <blockSize 2^k (bytes): say k> <stopwordsFile>\n", argv[0]);
			exit(0);
		}
	
		// Reads params (input file, output basename, and stopwords file)
		infile = argv[1];
		outbasename = argv[2];
		ksample = atoi(argv[3]);
		blocksize = atoi(argv[4]);
		blocksize = 1<<blocksize;
		stopwordsfile = argv[5];
		
//		finsize= fileSize(infile);
//		
//		if (! finsize) {
//			printf( "\nFILE EMPTY OR FILE NOT FOUND %s !!\nSkipping processement ...\n",infile);
//			exit(0);
//		}	
//	
//		// Opening the input text file.
//		if( (f_in = open(infile, O_RDONLY)) < 0) {
//			printf("Cannot read file %s\n", infile);
//			exit(0);
//		}	
//		inputBuffer = (byte *) malloc(finsize *sizeof(byte));// +1);
//		read (f_in,inputBuffer,finsize);	
//		close (f_in);	
		
		
	{
		//printf("\n parametros <<%s>>\n\n",stopwordsfile);	
		//build_index (inputBuffer, finsize, stopwordsfile, &Index);  /** building the index */

		fprintf(stderr,"\n basename = %s, outname = %s, blocksize = %d, sampleK = %d stopwords= %s\n",infile, outbasename, blocksize,ksample, stopwordsfile);
		build_index_baseFile((uchar *)infile, blocksize,  ksample, stopwordsfile,  &Index);

		/** saving the index to disk*/
		save_index (Index, outbasename);

		/** tells the mem used by the index */
		ulong indexsize;		
		index_size(Index, &indexsize);
		fprintf(stderr,"\nIndex occupied %d bytes (in memory)",indexsize);


//		void *Index2;																				
//		load_index(outbasename, &Index2);	
//		cmpindexes(Index,Index2);																			
//		free_index(Index2);

		/** freeing the index */		 			
		free_index(Index);	 			
		
	fprintf(stderr,"\n\n");
		
		//**loading the index2 */
		void *Index2;
		load_index(outbasename, &Index2);
		//Index2=Index;

		// SEARCHING FOR A TEXT PATTERN (word/phrase).
		{unsigned char textPattern[MAX_TEXT_PATTERN_SIZE];
		 int occ;
		 int len;
		 uint *occs;
		 int i;
		 printf("\nSEARCH TEST for LOCATE\n");
			while(1) {	
				printf("Intro string: ");
				fgets((char*)textPattern, MAX_TEXT_PATTERN_SIZE, stdin);
				len = strlen((char*)textPattern);
				if (!strcmp((char*)textPattern,"\n") ) break;
				textPattern[len-1] = '\0';
				len --;
				
				locate(Index2, textPattern, len, (ulong **)&occs, (ulong *)&occ);
				
				printf("\n*** %s occurs %d times: In the source text in positions:\n\t",textPattern,occ);
				for (i=0;i<occ;i++) 
					printf("[%u]",occs[i]);
				fflush(stderr);	
				if (occ) free(occs);		
							
			if (!strcmp((char*)textPattern,"\n") ) break;
			}
		}	
		
		
		//printf("BCS.SIZE = %d",((twcsa*)Index2)->occs.bcsSize);
		
		// DISPLAYING THE OCCURRENCES OF A TEXT PATTERN (word/phrase).
		{unsigned char textPattern[MAX_TEXT_PATTERN_SIZE];
		 int occ;
		 printf("\nSEARCH TEST for DISPLAY (non pizzachili)\n");
			while(1) {	
				printf("Intro string: ");
				fgets((char*)textPattern, MAX_TEXT_PATTERN_SIZE, stdin);
				if (!strcmp((char*)textPattern,"\n") ) break;
				textPattern[strlen((char*)textPattern)-1] = '\0';
				occ = displayTextOcurrences(Index2,textPattern,strlen((char*)textPattern),10);
				printf("Ocurrences = %d\n", occ);
				
				if (!strcmp((char*)textPattern,"\n") ) break;
			}
		}		
		

			free_index(Index2);	
	}
	
//checkSimpleSort();

}
	
#endif




/**------------------------------------------------------------------ 
  *  MAIN PROGRAM: FOR QUERIING THE INDEX
  *------------------------------------------------------------------ */
#ifdef SEARCHERFACADEWITHMAIN
#define COUNT 		('C')
#define LOCATE 		('L')
#define DISPLAY 	('D')


double getTime (void)
{
	double usertime, systime;
	struct rusage usage;

	getrusage (RUSAGE_SELF, &usage);
	usertime = (double) usage.ru_utime.tv_sec +
		(double) usage.ru_utime.tv_usec / 1000000.0;
	systime = (double) usage.ru_stime.tv_sec +
		(double) usage.ru_stime.tv_usec / 1000000.0;
	//return (usertime + systime);
	return (usertime );
}

void usage(char * progname) {	
	fprintf(stderr, "\nThe program loads <index> and then executes over it the\n");
	fprintf(stderr, "queries it receives from the standard input. The standard\n");
	fprintf(stderr, "input comes in the format of the files written by \n");
	fprintf(stderr, "genpatterns or genintervals.\n");
	fprintf(stderr, "%s reports on the standard error time statistics\n", progname);
	fprintf(stderr, "regarding to running the queries.\n\n");
	fprintf(stderr, "Usage:  %s <index> [<type> [length] [wordsbefore][V] ]\n", progname);
	fprintf(stderr, "\n\t<type>   denotes the type of queries:\n");
	fprintf(stderr, "\t         %c counting queries;\n", COUNT);
	fprintf(stderr, "\t         %c locating queries;\n", LOCATE);
	fprintf(stderr, "\t         %c displaying queries;\n", DISPLAY);
	fprintf(stderr, "\n\t[length] must be provided in case of displaying queries (D)\n");
	fprintf(stderr, "\t         and denotes the number of characters to display\n");
	fprintf(stderr, "\t         before and after each pattern occurrence.\n");
	fprintf(stderr, "\n\t[V]      with this options it reports on the standard output\n");
	fprintf(stderr, "\t         the results of the queries. The results file should be\n");
	fprintf(stderr, "\t         compared with trusted one by compare program.\n\n");
}

void readNumPatterns (ulong *numPhrase, ulong *numOfWords);
uchar *inputWord( ulong *len);
void loadPatterns(uchar ***patterns,ulong **lens, ulong *numpatt);

void do_test (void *Index);
void do_locate  (void *Index);
void do_display (void *Index, ulong numWords );

static double Load_time;
static ulong Index_size, Text_length;

FILE * timeFile;

	/**************************************/
	int main(int argc, char* argv[])
	{
		char *infile;	// baseName for the index files
		char query_type;
		void *Index;
		int error = 0;
		
		printf("\n*Searching a block-addressing index with canonical words ");
		printf("\n*CopyRight (c) 2007 [LBD & G.N.]\n\n");
	
		// Reads input parameters from command line.
		if(argc < 2) {
			printf("Use: %s <index basename> [<L|D> [words-snippet] ] \n", argv[0]);
			exit(0);
		}
	
		// Reads params (input file, query_type
		infile = argv[1];
		

char tmpname[1000];
sprintf(tmpname,"timesIIblocksCanonicals[%c].txt",*argv[2]);

		//**loading the index */	
		double atime = getTime();
		load_index(infile, &Index);
/**/    timeFile = fopen(tmpname,"a");
/**/    fprintf(timeFile,"\n====================================================");
/**/    fprintf(timeFile,"\n Search II-blocks with canonicals: %s",argv[1]);
		

// 	 twcsa *wcsa = (twcsa *) Index;
// 	 sampleK = wcsa->occs.sampleK;
// 	 bucksize = wcsa->occs.occsBucketSize;
// 	 absBits = wcsa->occs.absOffsetSize;
// 	 bcBits = wcsa->occs.bcOffsetSize;
// 	 zoneoccs = wcsa->occs.zoneOccs;
// 	 bcs = wcsa->occs.bcs;
		printf("\n load time = %f sec", getTime() -atime);
		ulong memSize;
		index_size(Index, &memSize);
		twcsa *wcsa = (twcsa*) Index;
		printf("\n index_size = %9d bytes = %6d kb, ksample= %3d , block_size= %d ",memSize,memSize/1024,wcsa->occs.sampleK, wcsa->blockSize);
		fprintf(timeFile,"\n index_size = %9d bytes = %6d kb, ksample= %3d , block_size= %d ",memSize,memSize/1024,wcsa->occs.sampleK, wcsa->blockSize);
		if(argc == 2) {
			do_test(Index);			
		}			
		else {		
			query_type = argv[2][0];	
	
			switch (query_type){
				case COUNT:			
					//do_count();
					break;
				case LOCATE:
					do_locate(Index);
					break;	
				case DISPLAY:
					if (argc < 4) {
						printf("Use: %s <index basename> [<L|D> [words-snippet] ] \n", argv[0]);	
						exit (1);
					}
					do_display(Index,(ulong) atol(argv[3]));
					break;
									
				default:
					fprintf (stderr, "Unknow option: main run\n");
					exit (1);
			}	
		}	
					
		free_index(Index);	
		fclose(timeFile);
	
	}

    /*********************************/
void do_test(void *Index) {
		// SEARCHING FOR A TEXT PATTERN (word/phrase).
		{unsigned char textPattern[MAX_TEXT_PATTERN_SIZE];
		 int occ;
		 int len;
		 uint *occs;
		 int i;
		 printf("\nSEARCH TEST for LOCATE\n");
			while(1) {	
				printf("Intro string: ");
				fgets((char*)textPattern, MAX_TEXT_PATTERN_SIZE, stdin);
				len = strlen((char*)textPattern);
				if (!strcmp((char*)textPattern,"\n") ) break;
				textPattern[len-1] = '\0';
				len --;
				
				locate(Index, textPattern, len, (ulong **)&occs, (ulong *)&occ);
				
				printf("\n*** %s occurs %d times: In the source text in positions:\n\t",textPattern,occ);
				for (i=0;i<occ;i++) 
					printf("[%u]",occs[i]);
				fflush(stderr);	
				if (occ) free(occs);		
							
			if (!strcmp((char*)textPattern,"\n") ) break;
			}
		}	
		
				
		// DISPLAYING THE OCCURRENCES OF A TEXT PATTERN (word/phrase).
		{unsigned char textPattern[MAX_TEXT_PATTERN_SIZE];
		 int occ;
		 printf("\nSEARCH TEST for DISPLAY (non pizzachili)\n");
			while(1) {	
				printf("Intro string: ");
				fgets((char*)textPattern, MAX_TEXT_PATTERN_SIZE, stdin);
				if (!strcmp((char*)textPattern,"\n") ) break;
				textPattern[strlen((char*)textPattern)-1] = '\0';
				occ = displayTextOcurrences(Index,textPattern,strlen((char*)textPattern),8);
				printf("Ocurrences = %d\n", occ);
				if (!strcmp((char*)textPattern,"\n") ) break;
			}
		}		
	}


	/*************************************************/
	void do_display(void *Index, ulong radixNumWords) {
		
		int error = 0;
		ulong numocc, length, i, tot_numcharext = 0;
		double time, tot_time = 0;
		
		uchar *pattern;
		ulong numpatt; 
		uchar ** patterns;
		ulong *lens;

		loadPatterns(&patterns, &lens, &numpatt);
		fflush(stdout);fflush(stderr);
		printf ("\nStarting display search over %d patterns, and radix = %d valid words ",numpatt,radixNumWords);
		
		time = getTime ();
		for (i=0; i<numpatt; i++) {
			printf("\n iter %3d of %3d",i,numpatt);	
			pattern=patterns[i];
			length = lens[i];
			
			/* Display */
			tot_numcharext+=displayTextOcurrencesNoShow(Index, pattern, length, radixNumWords,10000);
					    	  			
		}

		tot_time = (getTime () - time);
		fflush(stdout);fflush(stderr);		
		for (i=0;i<numpatt; i++) {		
			free(patterns[i]);   
		}
		free(patterns);
		free(lens);
	
		fprintf (stderr, "Total num chars extracted = %lu\n", tot_numcharext);
		fprintf (stderr, "\nDisplay time = %.4f secs\n\n", tot_time);
		fprintf (stderr, "Time_display/Tot_num_chars = %.4f msecs/chars\n\n", (tot_time*1000) / tot_numcharext);
		fprintf (timeFile, "Total num chars extracted = %lu\n", tot_numcharext);
		fprintf (timeFile, "\nDisplay time = %.4f secs\n\n", tot_time);
		fprintf (timeFile, "Time_display/Tot_num_chars = %.4f msecs/chars\n\n", (tot_time*1000) / tot_numcharext);
		//fprintf (stderr, "(Load_time+Time_display)/Tot_num_chars = %.4f msecs/chars\n\n", ((Load_time+tot_time)*1000) / tot_numcharext);
		
	}

	/*************************************************/
	void
	do_locate (void *Index)
	{
		int error = 0;
		ulong numocc, length, *occ, tot_numocc = 0;
		
		double time, tot_time = 0;
		
		ulong i;
	
		uchar *pattern;
		ulong numpatt; 
		uchar ** patterns;
		ulong *lens;
		
		loadPatterns(&patterns, &lens, &numpatt);

		fflush(stdout);fflush(stderr);
		printf ("\nStarting locate search over %d patterns:",numpatt);
		
		
		time = getTime ();
		//printf("\n iter %4d /     ",numpatt);
		for (i=0; i<numpatt; i++) {	
			pattern=patterns[i];
			length = lens[i];			
			//printf("\n iter %4d of %d [pattern = %s]",i,numpatt,pattern);
			//printf("\b\b\b\b%4d",i);
			/* Locate */
			error =	locate (Index, pattern, length, &occ, &numocc);
			printf("\n iter %4d of %4d , occs=%6d,  <<pattern = %s>>",i,numpatt,numocc,pattern);
		
			tot_numocc += numocc;
			
			if (numocc) free (occ);				
		}
	
		tot_time += (getTime () - time);
	
		fflush(stdout);fflush(stderr);
		//free patterns and lens arrays.
		for (i=0;i<numpatt; i++) {		
			free(patterns[i]);   
		}
		free(patterns);
		free(lens);

/**/ fprintf (timeFile, "\tTotal Num occs found = %lu \n", tot_numocc);
/**/ fprintf (timeFile, "\tLocate time = %.4f secs\n", tot_time);
/**/ fprintf (timeFile, "\tLocate_time/Num_occs = %.4f msec/occss\n", (tot_time * 1000) / tot_numocc);
/**/ fprintf (timeFile, "\t\t(Load_time+Locate_time)/Num_occs = %.4f msecs/occs\n\n", ((tot_time+Load_time) * 1000) / tot_numocc);
	
		//Show stats.
		fprintf (stderr, "Total Num occs found = %lu\n", tot_numocc);
		fprintf (stderr, "Locate time = %.4f secs\n", tot_time);
		fprintf (stderr, "Locate_time/Num_occs = %.4f msec/occss\n\n", (tot_time * 1000) / tot_numocc);
		fprintf (stderr, "(Load_time+Locate_time)/Num_occs = %.4f msecs/occs\n\n", ((tot_time+Load_time) * 1000) / tot_numocc);
	
	}




/*************************************************/
void readNumPatterns (ulong *numPhrase, ulong *numOfWords) {
	fscanf(stdin,"#numberOfFrases::%9d, #NumberOfWordsPerPhrase:: %9d\n",numPhrase, numOfWords);
	//fprintf(stderr," numOfFrase = %9d and numberOfWordsPerPhrase = %9d",*numPhrase, *numOfWords);
}

/*************************************************/	
uchar *inputWord( ulong *len) {
	ulong tmp; //skips one integer.
	uchar *phrase;
	
	fscanf(stdin,"%9d %9d ", &tmp, len);	
	phrase = (uchar * ) malloc (sizeof(char) * ((*len)+1));
	fread(phrase,sizeof(uchar), *len,stdin);
	phrase[*len] = '\0';
	//fscanf(stdin,"\n");
	
	return phrase;	
}

/*************************************************/
void loadPatterns(uchar ***patterns,ulong **lens, ulong *numpatt) {
	uchar *pattern;
	ulong i;
	ulong   wordsPerPhrase;   //number of patterns and number of words per phrase.	
	readNumPatterns(numpatt,&wordsPerPhrase);  //number of patterns and number of words per phrase.
	
	
	uchar **pats;
	ulong length, *thelens;
	pats = (uchar **) malloc (sizeof(uchar *) * (*numpatt));
	thelens = (ulong *) malloc (sizeof(ulong) * (*numpatt));
	
	for (i=0;i< (*numpatt);i++) {
		pats[i] = inputWord(&length);
		thelens[i]= length;
	}
	*patterns = pats;
	*lens=thelens; 
}
	
#endif
