/* Multipattern random searches using set Horspool's algorithm adapted to 
search inside a text compressed with End-Tagged Dense Code.
Lightweight Natural Language Text Compression: Information Retrieval 2006
Copyright (C) 2005 Antonio Fariña.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Author's contact: Antonio Fariña, Dept. of Computer Science, University of
A Coruña. Campus de Elviña s/n. Spain  fari@udc.es
*/

/*-----------------------------------------------------------------------
  search.c:     Implements a trie and deals with the search of a set of patterns
  inside a compressed file (searchFullFile()). Searching the whole file is done 
  one buffer at a time.
 ------------------------------------------------------------------------*/

#ifndef HORSPOOLINCLUDED
#define HORSPOOLINCLUDED

#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <malloc.h>
#include <stdio.h>


 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
       
#include "basics.h"
#include "dataTypesFacade.h"
#include "bc.h"
#include "ii.h"
#include "filterWord.h"

//#define MIN(a,b) (a < b) ? a : b
//#define MAX(a,b) (a > b) ? a : b
//
//#define LINE_UP(x,numBytes) ((x>>(4-numBytes)*8))   //move "4-bytes" bytes to the right
//						   						   	//so we can write "bytes" bytes from &x
//#ifdef LITTLE_ENDIAN
//	#define LE_LONG32(x)  (((x&0xff000000)>>24)|((x&0x00ff0000)>>8)|((x&0x0000ff00)<<8)|((x&0x000000ff)<<24))
//#else
//	#define LE_LONG32(x) x
//#endif


//#define MAX_STATES 2048
//#define MAX_STATES 4096
//#define NORMAL 4096         

#define MAX_CODEWORD_SIZE 4

#define MAX_STATES 4096
#define NORMAL 8192         //It must be an integer int larger than "NUMBER OF PATTERNS""
#define ROOT_STATE  0		//ROOT STATE OF THE TRIE.

//#define MAX_STATES 16383
//#define NORMAL 16384
#define INFINITE 100000


	//stores information about each search pattern
	struct t_patt {
		//byte word[MAX_CODEWORD_SIZE];
		byte *word;
		unsigned int size;
		unsigned long matches;
		unsigned int flag;
	};
	typedef struct t_patt t_pattern;

	struct t_pats {        //a structure containing the list of searched patterns
		t_pattern *patterns;
		unsigned int size;
		unsigned int lmin;
		unsigned int lmax;
	};
	typedef struct t_pats *t_patterns;


	//Definition of the trie data structure.
	struct tnod {
		unsigned int n;  //number of patterns this node occurs in
		unsigned int f;  //type of node (and reference to the vector of patterns		
	};
	typedef struct tnod tnodetrie;

	struct t_ttrie {
		tnodetrie nodes[MAX_STATES];         		 //nodes of the trie
		unsigned int transitions[MAX_STATES][256];   //transitions for each node of the tree, with each character read.
		unsigned long zeroNode;		//number of states of the trie used.
		uint stack;								// holds all the free states of the trie: initialy from 1->NORMAL-1
	};
	typedef struct t_ttrie *t_trie ;                //Our "trie" data type





/*********************************************/
	//t_patterns initializePatterns(int numpatts, byte **patts, int *lens);
	t_patterns initilializePatternsFromCodes (byte *codes, uint numBytesCodes) ;
	void inicialize_d(t_patterns P, byte *d);
	void updateDVector(int newSize, byte *oldPattern, int oldSize, t_patterns P, byte *d);
	void updateDVectorByAddition(int newSize, t_patterns P, byte *d);
	t_trie inicializeTrie(t_patterns P);
	void printTrie(t_trie trie,int maxstate);
	void removeFromTrie(t_trie trie, byte *code, int size);
	void addToTrie(t_trie trie, byte *nuevo, int size, unsigned int flag);
	uint searchInBufferETDC (byte *d, uint startOffset, byte *begin, byte *end, uint minLen, t_trie trie )	;

	void searchFullFile (char *infile, int numWords, byte **words, int *lens);

	int matchLeft(uint pos, uint *IDs, uint numIDs, uint pivot, twcsa *wcsa, uint *returnedPos); 
	int matchRight(uint pos, uint *IDs, uint numIDs, uint pivot, twcsa *wcsa, uint *returnedPos);

	uint searchPositionsFullFile (byte *d, uint startOffset, byte *begin, byte *end, uint minLen,
                      t_trie trie, twcsa *wcsa,uint *IDs, uint numIDs, uint pivot, uint *positions,
                       uint *size, uint rightFirst,int (*f1) (uint, uint*, uint, uint, twcsa *, uint *),
                       int (*f2) (uint, uint*, uint, uint, twcsa *, uint *));




#endif

