#include "ii.h"

	/**------------------------------------------------------------------
	Compares two slots (by frequency). For qsort of variants && separators
	------------------------------------------------------------------ */			   
	int qsortCompareUintAsc(const void *arg1, const void *arg2) {
		tUintSort *a1 = (tUintSort *) arg1;
		tUintSort *a2 = (tUintSort *) arg2;
		return a1->value >= a2->value;
	}	

/** ------------------------------------------------------------------
  * loads the compressed text (ETDC) into memory
  * ------------------------------------------------------------------*/
int loadTextInMem (byte **text, uint *size, char *basename) {
	int file;
	uint sizeFile;
	byte *buffer;

	char *filename;	
	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));
	strcpy(filename, basename); strcat(filename, ".");	strcat(filename, ETDCDATA);

	sizeFile= fileSize(filename);
	buffer = (byte *) malloc (sizeof(byte) * sizeFile);

	//fprintf(stderr,"\tLoading Compressed Text into memory from %s\n", filename);
	if( (file = open(filename, O_RDONLY)) < 0) {
		printf("Cannot open file %s\n", filename);
		exit(0);
	}
	read(file, buffer, sizeFile);
	close(file);			
	*text = buffer;
	*size = sizeFile;
	
	free(filename);
	return 0;
}


/** ------------------------------------------------------------------
  * saves the compressed text (ETDC) to disk
  * ------------------------------------------------------------------*/
int saveTextInMem (byte *text, uint size, char *basename) {
	int file;
	uint sizeFile;
	byte *buffer;

	char *filename;	
	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));
	strcpy(filename, basename); strcat(filename, ".");	strcat(filename, ETDCDATA);
	unlink(filename);
	//fprintf(stderr,"\tSaving Compressed Text (ETDC) into disk  %s\n", filename);
	{		
		if( (file = open(filename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}	
		write(file, text, sizeof(byte) * size);		
		close(file);		
	}	
	
	free(filename);
	return 0;
}


/**------------------------------------------------------------------
 Loads the VOCABULARY of ETDC and sets also the pointers from
 the byte *words array.
 ----------------------------------------------------------------- */
int loadVocabularyETDC (void *index, char* basename){
	twcsa *wcsa=(twcsa *) index;	
	uint nwords;
	byte *zoneMem;
	char *filename;	
	uint sizeFile;
	int file;
	ulong i;
	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));
	
	{	
		strcpy(filename, basename);
		strcat(filename, ".");
		strcat(filename, ETDCVOC);
		sizeFile= fileSize(filename)-sizeof(uint);

		//fprintf(stderr,"\tLoading vocabulary of ETDC from index %s\n", filename);

		if( (file = open(filename, O_RDONLY)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}		
		
		//the number of canonical words
		read(file, &nwords, sizeof(uint));
		wcsa->nwords = nwords;
		wcsa->words = (byte **) malloc(sizeof(byte *) * (wcsa->nwords)); 
		fprintf(stderr,"Vocabulary: %d words\n",nwords);
		wcsa->wordsZone.size = sizeFile;
		wcsa->wordsZone.zone = (byte *) malloc (sizeof(byte) * sizeFile);
		read(file,wcsa->wordsZone.zone, (sizeof(byte) * sizeFile));
		//printf("\n vocabulary of ETDC:: words loaded sucessfuly..");fflush(stdout);
				
		//setting the pointers in byte **words vector
		zoneMem = wcsa->wordsZone.zone;
		for(i = 0; i < (wcsa->nwords); i++) {
			wcsa->words[i]=zoneMem;
			while (1) {if (!(*zoneMem++)) break;} //reads word
		}
		close(file);
	
	}		
	free(filename);
//
//	
//	{	uint i,j; 
//		for(i = 0; i <(wcsa->nwords); i++) {
//			fprintf(stderr,"\n word[%d]= <<%s>>", i, wcsa->words[i]);			
//			//j = wtinHashTable(wcsa->hash, wcsa->words[i], strlen((char *)wcsa->words[i]), &addrInTH);
//			//if (j) {fprintf(stderr," id = %d",  wcsa->hash->hash[addrInTH].id);}
//		}
//	}
	
	return (0);
} 


/** ------------------------------------------------------------------
  * Saves the VOCABULARY of ETDC to disk
  * ------------------------------------------------------------------*/

int saveVocabularyETDC (void *index, char* basename) {
	int file;

	char *filename;	
	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));
	strcpy(filename, basename); strcat(filename, ".");	strcat(filename, ETDCVOC);
	unlink(filename);
	
	twcsa *wcsa=(twcsa *) index;	
	uint nwords;
	{		
		unlink(filename);
		//fprintf(stderr,"\tSaving Vocabulary (ETDC) into disk %s\n", filename);		
		if( (file = open(filename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}	

		write(file, &(wcsa->nwords), sizeof(uint));		
		write(file, wcsa->wordsZone.zone, wcsa->wordsZone.size * sizeof(byte));		
		close(file);		
	}	
	
	free(filename);
	return 0;
}






	/**------------------------------------------------------------------
	Compares two slots (alphanumericaly). For qsort of canonical words 
	------------------------------------------------------------------ */			   
	int canonicalWordsCompareAlpha(const void *arg1, const void *arg2) {
		tauxslots *a1 = (tauxslots *) arg1;
		tauxslots *a2 = (tauxslots *) arg2;
		return strcmp((char*)a1->word, (char *)a2->word);	
	}


/** ------------------------------------------------------------------
  * Creates a temporal hash table with the canonical words and computes number of variants.
  * stores hash->hash[i].word == canonical word of size hash->hash[i].len
  * stores hash->hash[i].numVars == number of variants of this canonical word.
  *        hash->hash[i].numVars == 0 if the slot is empty. 
  * ------------------------------------------------------------------*/
void generateTempHashOfCanonicals(Wt_hash *canonicalsHash, void *index) {
	twcsa *wcsa=(twcsa *) index;
	tCanonical canonical;
	byte *aWord;
	ulong size,i,pos,addrInTH, n;
	Wt_hash wordshash;
	
	wordshash = wtinitialize_hash(wcsa->nwords, wcsa->wordsZone.size); //creates the hash table & zoneMem
	*canonicalsHash =wordshash;
	
	n=wcsa->nwords;
	for (i=0;	i<n;i++) {		
		aWord=wcsa->words[i];
		size = strlen((char*)aWord);	
		
		if (isIndexedWord(wcsa,i)) {
			getCanonical(&canonical, aWord, size, wcsa->stopwordshash);
			aWord = canonical.word;
			size  = canonical.size;

			//updating the canonical word (inserting or adding numVars)
			pos = wtsearch (wordshash, aWord, size, &addrInTH);
			//fprintf(stderr,"Insertando %s, con %d vars\n",aWord,wordshash->hash[addrInTH].numVars);
			if (wordshash->hash[addrInTH].numVars == 0){
				wtinsertElement (wordshash, aWord, size, &addrInTH);  //adds the canonical word						
			}
			wordshash->hash[addrInTH].numVars ++;  // so numVars > 0;
		}	
	}
}


/*Calculates the array of sorted canonical words (and: byte **canonicals, byte *zoneCanonicals and ncanonicals*/
void getSortedCanonicalsArray(Wt_hash canonicalsHash, void *index) {
	twcsa *wcsa=(twcsa *) index;	

	int i;
	uint ncanonicals =0;
	uint lenCanonicalsZone =0;      
	byte *aWord;            
	ulong size,pos,addrInTH;
	
	//** count the number of canonical Words and sets wcsa->ncanonicals.
	//** counts also the number of bytes needed to allocate canonical words (in zoneCanonicals.zone).
	for (i=0; i<canonicalsHash->SIZE_HASH;i++) {
		if(canonicalsHash->hash[i].numVars !=0) {
			//printf("\n [%d] word in the hashtable = %s, lend =%d vars_= %5d",count,canonicalsHash->hash[i].word,canonicalsHash->hash[i].len,canonicalsHash->hash[i].numVars);
			ncanonicals ++;
			lenCanonicalsZone += canonicalsHash->hash[i].len +1 ;
		}					
	}			
	
	printf("\n [%5d]:: lencanoincalsZone and ncanonicals= %d ",lenCanonicalsZone,ncanonicals);
	wcsa->ncanonicals = ncanonicals; //number of different canonical words.
	wcsa->canonicalsZone.size = lenCanonicalsZone;
	wcsa->canonicalsZone.zone = (byte *) malloc (sizeof(byte) * lenCanonicalsZone);	
	wcsa->canonicals = (byte **) malloc(sizeof(byte *) * (wcsa->ncanonicals)); 
	

	{/**Sorting **/
		//copying to auxiliar structure for sorting
		tauxslots *auxcan = (tauxslots *) malloc (sizeof(tauxslots) * (ncanonicals));			
		pos=0;
		for (i=0; i<canonicalsHash->SIZE_HASH;i++) {
			if(canonicalsHash->hash[i].numVars !=0) {
				auxcan[pos].word = canonicalsHash->hash[i].word;
				auxcan[pos].slot= i;
				pos++;
			}					
		}		

		
		//sorting
		qsort(auxcan, ncanonicals, sizeof(tauxslots), canonicalWordsCompareAlpha);	
									
		//creating canonicals from the sorted auxiliar structure.
		{	byte *dst = wcsa->canonicalsZone.zone;
			for (i=0; i<ncanonicals; i++) {
				wcsa->canonicals[i] = dst;
				//fprintf(stderr,"Canonical[%d]: %s\n",i,auxcan[i].word);;
				aWord = auxcan[i].word;
				while (*aWord) {*dst = *aWord; dst++; aWord++;}
				*dst=0; dst++;
			}
		}
		
//		fprintf(stderr,"Canonical: %d\n",ncanonicals);

		free(auxcan);
	}			
	//for (i=0;i<ncanonicals;i++) printf("\n [%5d]:: canonical : %s ",i, wcsa->canonicals[i]);
}

/**------------------------------------------------------------------
   Loads the Ascii of the canonical words from disk into wcsa
 ----------------------------------------------------------------- */
int loadCanonicalsAscii (void *index, char* basename){
	twcsa *wcsa=(twcsa *) index;	
	uint ncanonicals;
	byte *zoneMem;
	char *filename;	
	uint sizeFile;
	int file;
	ulong i;
	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));
	
	{	
		strcpy(filename, basename);
		strcat(filename, ".");
		strcat(filename, CANONICALS_EXT);
		sizeFile= fileSize(filename)-sizeof(uint);

		//fprintf(stderr,"\tLoading Ascii text of canonicals from index %s\n", filename);

		if( (file = open(filename, O_RDONLY)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}		
		
		//the number of canonical words
		read(file, &ncanonicals, sizeof(uint));
		wcsa->ncanonicals = ncanonicals;
		wcsa->canonicals = (byte **) malloc(sizeof(byte *) * (wcsa->ncanonicals)); 

		wcsa->canonicalsZone.size = sizeFile; // 4 bytes less than the actual file-size.
		wcsa->canonicalsZone.zone = (byte *) malloc (sizeof(byte) * sizeFile);
		read(file,wcsa->canonicalsZone.zone, (sizeof(byte) * sizeFile));
		//printf("\n canonicals ASCII:: words loaded sucessfuly..");fflush(stdout);
				
		//setting the pointers in byte **words vector
		zoneMem = wcsa->canonicalsZone.zone;
		for(i = 0; i < (wcsa->ncanonicals); i++) {
			wcsa->canonicals[i]=zoneMem;
			//while (1) {if (!(*zoneMem++)) break;} //reads word
			while (*zoneMem) {
				zoneMem++;					
			} //skips current word
			zoneMem++;

		}
		close(file);
	
	}		
	free(filename);	
	return (0);
} 




/** ------------------------------------------------------------------
  * Saves the Ascii of the canonical words into disk
  * ------------------------------------------------------------------*/
int saveCanonicalsAscii (void *index, char* basename) {
	int file;

	char *filename;	
	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));
	strcpy(filename, basename); strcat(filename, ".");	strcat(filename, CANONICALS_EXT);
	unlink(filename);
	
	twcsa *wcsa=(twcsa *) index;	
	uint ncanonicals;
	{		
		unlink(filename);
		//fprintf(stderr,"\tSaving ascii text of canonicals into disk %s\n", filename);		
		if( (file = open(filename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}	

		write(file, &(wcsa->ncanonicals), sizeof(uint));		
		write(file, wcsa->canonicalsZone.zone, wcsa->canonicalsZone.size * sizeof(byte));		
		close(file);		
	}	
	
	free(filename);
	return 0;
}


/*************************************************************************************************/
/*************************************************************************************************/
/*************************************************************************************************/

/** ------------------------------------------------------------------
  * Encodes, for each canonical word "i", the positions of its variants
  * in the array "words". So that each canonical keeps the "ETDC-codes" 
  * associated to its variants.
  * ------------------------------------------------------------------*/
int encodePosOfVariantsOfCanonical (void *index, uint **tmpPosVariants, uint *tmpNumVariants) {
	twcsa *wcsa=(twcsa *) index;	
	register uint i;
	register uint ncanonicals = wcsa->ncanonicals;
	register uint nwords = wcsa->nwords;
	
	byte *varsETDC; 			//the ids of the variants encoded with ETDC.
	uint varsETDCSize; 		//number of bytes used by varsETDC
	uint *varsCanonicals; //indicates de offset in "varsETDC" where the variants of any canonical begin.
	
	//calculates the lenght of byte * varsETDC
	{	
		int x;
		varsETDCSize=0;	
		for (i=0;i<nwords;i++) {
			if (isIndexedWord(wcsa,i)) {				
				varsETDCSize+=getLenETDC(i);
			}
		}
	}
		
	//allocates memory
	//varsETDCSize = getAcumSizeCodesETDCFreq1(nwords);  
	varsETDC = (byte *) malloc (sizeof(byte) * varsETDCSize);
	varsCanonicals = (uint *) malloc (sizeof(uint) * (ncanonicals+1));

	//encodes the positions from tmpPosVariants
	{
		ulong code,k,pos;
		byte *dst;
		uint subPos=0;
		uint j;
		for (i=0;i<ncanonicals;i++) {
			varsCanonicals[i]=subPos;
			for (j=0;j<tmpNumVariants[i];j++) {
				pos = tmpPosVariants[i][j];
				getCode(pos,&code, &k);
				dst = &(varsETDC[subPos]);
				getBytes (dst,code,k);                    
				subPos+=k;
			}			
		}
		varsCanonicals[ncanonicals]=subPos;		
	}//encoding finishes.
	
									//checks it works ;)
									//	{
									//		uint j;
									//	 	uint pos, endpos;
									//	 	uint id;
									//		for (i=0;i<ncanonicals;i++) {
									//			pos   =varsCanonicals[i];
									//			endpos=varsCanonicals[i+1];
									//			printf("\n **[%5d]Canonical= %s** %d variants ==>",i,wcsa->canonicals[i], tmpNumVariants[i]);
									//			j=0;
									//			while (pos < endpos) {
									//				pos = decode (varsETDC,pos, &id);
									//				if (tmpPosVariants[i][j] !=  id) {printf("\n DIFFERENTS IDS !! VARSETDC encode|decode fails");exit(0);}
									//				printf("<<%s>>",wcsa->words[id]);
									//				j++;
									//			} 
									//		}
									//		printf("\n ** VARSETDC encoded && decoded sucessfully. Congratulations!!\n");
									//	}

	wcsa->varsCanonicals = varsCanonicals;
	wcsa->varsETDCSize 	 = varsETDCSize;
	wcsa->varsETDC 			 = varsETDC;

}

/** ------------------------------------------------------------------------------
  * Saves the file with the variants of the canonical :: varsETDC y varsCanonicals
  * ------------------------------------------------------------------------------*/
int saveVariantsCanonicals (void *index, char* basename) {
	int file;

	char *filename;	
	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));
	strcpy(filename, basename); strcat(filename, ".");	strcat(filename, VARIANTS_EXT);
	unlink(filename);
	
	twcsa *wcsa=(twcsa *) index;	
	uint ncanonicals;
	{		
		unlink(filename);
		//fprintf(stderr,"\tSaving varsStructure for the canonicals. into disk %s\n", filename);		
		if( (file = open(filename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}	

		write(file, &(wcsa->varsETDCSize), sizeof(uint));		   //the size of the ids compressed with ETDC
		write(file, wcsa->varsETDC, (wcsa->varsETDCSize) * sizeof(byte));		// the ids of the variants compressed with ETDC
		write(file, wcsa->varsCanonicals, (wcsa->ncanonicals +1) * sizeof(uint));		//the offset into varsETDC to each canonical word.
		
		close(file);		
	}	
	
	free(filename);
	return 0;
}

/**------------------------------------------------------------------
   Loads the Ascii of the canonical words from disk into wcsa
 ----------------------------------------------------------------- */
int loadVariantsCanonicals (void *index, char* basename){
	twcsa *wcsa=(twcsa *) index;	
	char *filename;	
	int file;
	ulong i;
	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));	
	{	
		strcpy(filename, basename);		strcat(filename, ".");		strcat(filename, VARIANTS_EXT);
		//fprintf(stderr,"\tLoading varsStructure for the canonicals. from disk %s\n", filename);

		if( (file = open(filename, O_RDONLY)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}		
		
		//the number of bytes of the ids of the variants encoded with ETDC
		read(file, &(wcsa->varsETDCSize), sizeof(uint));		   //the size of the ids compressed with ETDC
		//allocates mem.
		wcsa->varsETDC = (byte *) malloc(sizeof(byte) * (wcsa->varsETDCSize)); 
		wcsa->varsCanonicals = (uint *) malloc (sizeof(uint) * (wcsa->ncanonicals+1));
		//reads data.
		read(file, wcsa->varsETDC, (wcsa->varsETDCSize) * sizeof(byte));		// the ids of the variants compressed with ETDC
		read(file, wcsa->varsCanonicals, (wcsa->ncanonicals +1) * sizeof(uint));		//the offset into varsETDC to each canonical word.

		//printf("\nvarsStructure for the canonicals... loaded sucessfuly..");fflush(stdout);				
		close(file);	
	}		
	free(filename);
	
	return (0);
} 


/*************************************************************************************************/
/*************************************************************************************************/


	/* checks if a word in the vocabulary of ETDC is either indexed or not */
	int isIndexedWord(twcsa *wcsa, uint rank) {
		uint value = bitget((wcsa->indexedWords),rank);
		return value;
	}
	
	/** ------------------------------------------------------------------
	  * Creates the array wcsa->indexedWords to allow using "isIndexedWord function".
	  * ------------------------------------------------------------------*/
	int markIndexedWords (void *index) {
		register uint i, size;	
		uint valid;
		byte *aWord;
		tCanonical canonical;	
		twcsa *wcsa=(twcsa *) index;	

		uint n = wcsa->nwords;
		uint sizeInInts = ((n+W-1)/W);
		uint *indexedWords = (uint *) malloc ( sizeof(uint) * sizeInInts ) ;
		wcsa->indexedWords = indexedWords;
		
		//initializes to "non indexed word == 0"
		for (i=0;	i<sizeInInts;i++) indexedWords[i]=0;		

		//checks if the word in the vocabulary is not a separator nor a stopword.
		for (i=0;	i<n;i++) {		
			aWord=wcsa->words[i];
			size = strlen((char*)aWord);	
			//printf("\n TO CANONIZE %20s, len = %2d",aWord,size);
			if (!_Valid[*aWord]) {
				valid=0;
			}
			else{// _Valid[*aWord] && 
				getCanonical(&canonical, aWord,size, wcsa->stopwordshash);	
				//printf("** CANONICAL %s stopword %d",canonical.word,canonical.type == IS_STOPWORD);
				valid = (!canonical.type == IS_STOPWORD);
			}
			
			if (valid) bitset(indexedWords,i);  //sets the word as indexed.
		}
		
		return 0;
	}


/** Saving indexedWords array */
int saveIndexedWordsVector(void *index, char *basename) {
	twcsa *wcsa=(twcsa *) index;	
	int file;
	uint n = wcsa->nwords;
	uint sizeInInts = ((n+W-1)/W);
	uint sizeFile = (sizeInInts * sizeof(uint));
	uint *buffer = wcsa->indexedWords;

	char *filename;	
	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));
	strcpy(filename, basename); strcat(filename, ".");	strcat(filename, VOC_MARKED_INDEXED_WORDS);

	//fprintf(stderr,"\tSaving Vector of marks of indexed words %s\n", filename);
	unlink(filename);
	if( (file = open(filename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
		printf("Cannot open file %s\n", filename);
		exit(0);
	}
	write(file, buffer, sizeFile);
	close(file);				
	free(filename);
	return 0;
}

/** Loading indexedWords array */
int loadIndexedWordsVector(void *index, char *basename) {
	twcsa *wcsa=(twcsa *) index;	
	int file;
	uint sizeFile;
	uint *buffer;

	char *filename;	
	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));
	strcpy(filename, basename); strcat(filename, ".");	strcat(filename, VOC_MARKED_INDEXED_WORDS);

	sizeFile= fileSize(filename);
	buffer = (uint *) malloc (sizeFile);

	//fprintf(stderr,"\tLoading Vector of marks of indexed words %s\n", filename);
	if( (file = open(filename, O_RDONLY)) < 0) {
		printf("Cannot open file %s\n", filename);
		exit(0);
	}
	read(file, buffer, sizeFile);
	close(file);			
	free(filename);

	wcsa->indexedWords = buffer;
	return 0;
}	







/** **********************************************************************/
/** Some functions used during index construction                        */
/** **********************************************************************/


/** ------------------------------------------------------------------
  * Computes the frequency of each canonical word
  *          && sets the number of valid words in the source text: Nwords
  * ------------------------------------------------------------------*/
int computeFreqs (void *index, int **mappingToCanonical) {
	uint i;	
	twcsa *wcsa=(twcsa *) index;	
	byte *text = wcsa->text;
	uint N = wcsa->textSize;
	uint n = wcsa->ncanonicals;
	uint totalfreq=0;
	uint pos;


	//initializes mapToCanonical Array.
	int *mapToCanonical = (int *) malloc (sizeof(int) * wcsa->nwords);
	*mappingToCanonical = mapToCanonical;
	for (i=0;i<wcsa->nwords;i++) mapToCanonical[i] = -1;   //-1 if not an indexed word.
	for (i=0;i<wcsa->nwords;i++) {
		if (isIndexedWord(wcsa,i)) 
			mapToCanonical[i] = parseWordToCanonicalID(wcsa, wcsa->words[i]);
	}

	//Computes the frequencies for each canonical word.
	//fprintf(stderr,"Computing frequencies of each word");	
	for (i=0;i<n;i++) wcsa->freqs[i]=0;
	
	//decode && increase freqs :: CALCULATES THE FREQUENCIES OF EACH CANONICAL WORD (sum of variants).
	i=0;
	uint num=0;
	while (i<N) {
		//BC_TO_DEC(text,i,num);
		i =decode (text, i, &num) ;
		
		if (mapToCanonical[num] !=-1){
			//if (isIndexedWord(wcsa,num)) {
			//pos = parseWordToCanonicalID(wcsa, wcsa->words[num] );
			pos =mapToCanonical[num];
			wcsa->freqs[pos]++;
			//fprintf(stderr,"\n decoded word with id = %d is ",num);fflush(stderr);
			//fprintf(stderr,"<<%s>> and canonical is <<%s>> ",wcsa->words[num], wcsa->canonicals[pos]); fflush(stderr);	
			totalfreq++;	
		}
	}		
	wcsa->Nwords = totalfreq;
	//fprintf(stderr,"--> freqs[0] = %d, freqs[%d] = %d",wcsa->freqs[0], n-1, wcsa->freqs[n-1]);	
	//fprintf(stderr,"\n Number of valid words in the source text is: %d",wcsa->Nwords);	
	return 0;
}

/** ------------------------------------------------------------------
  * Creates the lists of occurrences of all the words
  * Block-oriented: Only 1 occurrence per block is marked for each word.
  * A word belongs to a block if its associated codeword starts in that block.
  * ------------------------------------------------------------------*/
int createListsOfOccurrences(void *index, uint ***occList, uint **lenList, int *mapToCanonical) {
	uint i;	
	twcsa *wcsa=(twcsa *) index;	
	byte *text = wcsa->text;
	uint N = wcsa->textSize;
	uint n = wcsa->ncanonicals;
	uint offset;
	uint blocksize = wcsa->blockSize;
	
	uint currblock = 0;
	uint posinblock=0;
	int canonicalID;

	//allocating memory.
	uint **occs;
	occs = (uint **) malloc (sizeof(uint *) * n);
	for (i=0;i<n;i++) occs[i]= (uint *) malloc (sizeof(uint) * wcsa->freqs[i]);
	
	//temporal array to keep the last index in the list of the i-th word
	uint *currpos = (uint *) malloc (sizeof(uint)*n);
	//temporal array to keep the block of the last occurrence of the i-th word
	int *lastBlock = (int *) malloc (sizeof(int)*n);
	for (i=0;i<n;i++) {currpos[i]=0; lastBlock[i] = -1;}
		
	//decode && increase freqs
	i=0;
	uint num=0;
	while (i<N) {
		offset =i/blocksize;  //optimizable if a power of 2 is used as block-size
		i = decode (text, i, &num);
		canonicalID = mapToCanonical[num];
		if (canonicalID != -1) {
			if (offset != lastBlock[canonicalID]) {  //only 1 occurrence per block is marked.
				occs[canonicalID][currpos[canonicalID]] = offset;  //word "num" occurrs in block "offset"
				currpos[canonicalID]++;			
				lastBlock[canonicalID] = offset;
			}	
		}
	}		
	free(lastBlock);	

	*occList = occs;	
	*lenList = currpos;
	return 0;
}



/** **********************************************************************/
/** Saving/Loading the index to/from disk                               **/
/** **********************************************************************/

/** Saving freq vector */
int saveFreqVector(uint *V, uint size, char *basename) {
	int file;
	uint sizeFile = size*sizeof(uint);
	uint *buffer = V;

	char *filename;	
	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));
	strcpy(filename, basename); strcat(filename, ".");	strcat(filename, FREQVECTOR);

	//fprintf(stderr,"\tSaving Vector of frequencies %s\n", filename);
	unlink(filename);
	if( (file = open(filename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
		printf("Cannot open file %s\n", filename);
		exit(0);
	}
	write(file, buffer, sizeFile);
	close(file);				
	free(filename);
	return 0;
}

/** Loading  freq vector */
int loadFreqVector(uint **V, uint *size, char *basename) {
	int file;
	uint sizeFile;
	uint *buffer;

	char *filename;	
	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));
	strcpy(filename, basename); strcat(filename, ".");	strcat(filename, FREQVECTOR);

	sizeFile= fileSize(filename);
	buffer = (uint *) malloc (sizeFile);

	//fprintf(stderr,"\tLoading Vector of frequencies %s\n", filename);
	if( (file = open(filename, O_RDONLY)) < 0) {
		printf("Cannot open file %s\n", filename);
		exit(0);
	}
	read(file, buffer, sizeFile);
	close(file);			
	free(filename);

	*V = buffer;
	*size = sizeFile/4;	
	return 0;
}


/** Saving lenList vector */
int savelenListVector(uint *V, uint size, char *basename) {
	int file;
	uint sizeFile = size*sizeof(uint);
	uint *buffer = V;

	char *filename;	
	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));
	strcpy(filename, basename); strcat(filename, ".");	strcat(filename, LENLISTS);

	//fprintf(stderr,"\tSaving Vector with lens for the list of each id: %s\n", filename);
	unlink(filename);
	if( (file = open(filename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
		printf("Cannot open file %s\n", filename);
		exit(0);
	}
	write(file, buffer, sizeFile);
	close(file);				
	free(filename);
	return 0;
}

/** Loading  lenList vector */
int loadlenListVector(uint **V, uint *size, char *basename) {
	int file;
	uint sizeFile;
	uint *buffer;

	char *filename;	
	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));
	strcpy(filename, basename); strcat(filename, ".");	strcat(filename, LENLISTS);

	sizeFile= fileSize(filename);
	buffer = (uint *) malloc (sizeFile);

	//fprintf(stderr,"\tLoading Vector with lens for the list of each id: %s\n", filename);
	if( (file = open(filename, O_RDONLY)) < 0) {
		printf("Cannot open file %s\n", filename);
		exit(0);
	}
	read(file, buffer, sizeFile);
	close(file);			
	free(filename);

	*V = buffer;
	*size = sizeFile/4;	
	return 0;
}

/** Saving the lists of occurrences of all the words */
int saveOccLists (void *index, char *basename) {
	int file;
	char *filename;	
	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));
	strcpy(filename, basename); strcat(filename, ".");	strcat(filename, LISTSOCC);
	//fprintf(stderr,"\tSaving the lists of occurrences %s\n", filename);
	unlink(filename);
			
	twcsa *wcsa=(twcsa *) index;	
	{		
		if( (file = open(filename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}	
	
		write(file, &(wcsa->blockSize), sizeof(uint));		
		write(file, &(wcsa->occs.sampleK), sizeof(uint));		
		write(file, &(wcsa->occs.zoneOccsSize), sizeof(uint)); //in # bits
		write(file, &(wcsa->occs.numBuckets), sizeof(uint));		
		write(file, &(wcsa->occs.absOffsetSize), sizeof(uint));		
		write(file, &(wcsa->occs.bcOffsetSize), sizeof(uint));		
		write(file, &(wcsa->occs.bcsSize), sizeof(uint));		

		//fprintf(stderr,"\n *sampleK = %d, zoneOccsSize = %d, numbucks= %d",wcsa->occs.sampleK, wcsa->occs.zoneOccsSize, wcsa->occs.numBuckets); 
		//fprintf(stderr,"\n *absoffsetSize = %d, bcOffsetSize = %d, occsBucketSize= %d",wcsa->occs.absOffsetSize, wcsa->occs.bcOffsetSize, wcsa->occs.occsBucketSize); 
		//fprintf(stderr,"\n *bcsSize = %d",wcsa->occs.bcsSize);

		write(file, wcsa->occs.occs, wcsa->ncanonicals * sizeof(uint));		
		write(file, wcsa->occs.zoneOccs, ((wcsa->occs.zoneOccsSize +W-1)/W) * sizeof(uint)); //zoneOccs kbitVec[bucketSize]
		write(file, wcsa->occs.bcs, wcsa->occs.bcsSize * sizeof(byte));		
		close(file);		
	}		
	free(filename);
	return 0;
}

/** Loading the lists of occurrences of all the words into "index" */
int loadOccLists (void *index, char *basename) {
	int file;
	char *filename;	
	twcsa *wcsa=(twcsa *) index;	

	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));
	strcpy(filename, basename); strcat(filename, ".");	strcat(filename, LISTSOCC);
	//fprintf(stderr,"\tLoading Lists of Occurrences %s\n", filename);
	{		
		if( (file = open(filename, O_RDONLY)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}	

		//some integers
		uint number;
		read(file, &(wcsa->blockSize), sizeof(uint));	
		read(file, &(wcsa->occs.sampleK), sizeof(uint));		
		read(file, &(wcsa->occs.zoneOccsSize), sizeof(uint)); //in # bits
		read(file, &(wcsa->occs.numBuckets), sizeof(uint));		
		read(file, &(wcsa->occs.absOffsetSize), sizeof(uint));		
		read(file, &(wcsa->occs.bcOffsetSize), sizeof(uint));	
		wcsa->occs.occsBucketSize = wcsa->occs.bcOffsetSize + wcsa->occs.absOffsetSize;
		read(file, &(wcsa->occs.bcsSize), sizeof(uint));		

		//printf("\n\t-sampleK = %d, blockSize = %d",wcsa->occs.sampleK, wcsa->blockSize); 

		//fprintf(stderr,"\n *sampleK = %d, zoneOccsSize = %d, numbucks= %d",wcsa->occs.sampleK, wcsa->occs.zoneOccsSize, wcsa->occs.numBuckets); 
		//fprintf(stderr,"\n *absoffsetSize = %d, bcOffsetSize = %d, occsBucketSize= %d",wcsa->occs.absOffsetSize, wcsa->occs.bcOffsetSize, wcsa->occs.occsBucketSize); 
		//fprintf(stderr,"\n *bcsSize = %d",wcsa->occs.bcsSize);
		
		//vector occs.
			wcsa->occs.occs = (uint *) malloc ( wcsa->ncanonicals * sizeof(uint) );		
		read(file, wcsa->occs.occs, wcsa->ncanonicals * sizeof(uint)); //wcsa->ncanonicals was loaded in "loadVocabulary"

		//vector zoneOccs (the kbitVector, with numBuckets of size (occsBucketSize) bits each
			wcsa->occs.zoneOccs = (uint *) malloc (((wcsa->occs.zoneOccsSize +W-1)/W) * sizeof(uint));
			wcsa->occs.zoneOccs[((wcsa->occs.zoneOccsSize +W-1)/W)-1] =0000; //so valgrind is shut up		
		read(file, wcsa->occs.zoneOccs, ((wcsa->occs.zoneOccsSize +W-1)/W) * sizeof(uint)); //zoneOccs kbitVec[bucketSize]

		//vector bcs.
			wcsa->occs.bcs = (byte *) malloc (wcsa->occs.bcsSize * sizeof(byte));		
		read(file, wcsa->occs.bcs, wcsa->occs.bcsSize * sizeof(byte));		
		close(file);		
	}		
	
	free(filename);
	return 0;
}



/** **********************************************************************/
/** Functions needed for searches                                       **/
/** **********************************************************************/

/** Uncompresses the list of occurrences of a given id-word **/
int uncompressList (void *index, uint id, uint **list, uint *len) {
	twcsa *wcsa=(twcsa *) index;	
	register uint absBits = wcsa->occs.absOffsetSize;
	register uint bcBits = wcsa->occs.bcOffsetSize;
	byte *bcs  = wcsa->occs.bcs;
	uint *zoneoccs = wcsa->occs.zoneOccs;
	uint lenlist  = wcsa->lenList[id];
	uint sampleK = wcsa->occs.sampleK;
	uint *alist = (uint *) malloc (sizeof(uint) * lenlist);

	//fprintf(stderr,"\nDecoding all the %d occurrences of word[%d]= <<%s>>",lenlist, id, wcsa->words[id]);fflush(stderr);		 		
	uint i, pos, bcpos, num, prevvalue;
	pos = wcsa->occs.occs[id]; //pos in zoneoccs
	for (i=0; i<lenlist;i++) {		
		if (! (i%sampleK)) {	//an abs sample
			prevvalue = bitread (zoneoccs, pos, absBits);
			alist[i]=prevvalue;
			pos+=absBits;
			bcpos = bitread (zoneoccs, pos, bcBits);
			pos +=bcBits;			
		}
		else { //decoding from bcs
			BC_TO_DEC(bcs, bcpos, num);
			prevvalue +=num;
			alist[i]=prevvalue;			
		}
	}		
	*list =alist;
	*len =i;	
//		fprintf(stderr,"\n[%d] decoded word at position text[%d] ",i,pos);
//		pos = decode (text, pos, &num) ;	
//		fprintf(stderr," and id = %d is ",num);fflush(stderr);
//		fprintf(stderr,"<<%s>>",wcsa->words[num]);fflush(stderr);
//		break;		 			
}	



	#include "iisvsETDCBlock.c"   //for the full text index (word-oriented with ETDC).




/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

/** !!! INCOMPLETE !!!, feito para atopar 1 erro !!! */

void cmpindexes (void *index1, void * index2) {
	uint i;
	twcsa *wcsa1 = (twcsa *) index1;
	twcsa *wcsa2 = (twcsa *) index2;
	fflush(stderr);
	fflush(stdout);
		
	fflush(stderr);fflush(stdout);
	printf("\n *******************COMPARING TWO INDEXES\n ");	
		
	if (wcsa1->nwords != wcsa2->nwords) {
		printf ("\n nowrds son distintos xxx <%d>, <%d>",wcsa1->nwords, wcsa2->nwords);fflush(stdout);return;
	}
		//printf ("\n nowrds son iguales xxx <%d>, <%d>",wcsa1->nwords, wcsa2->nwords);
	if (wcsa1->ncanonicals != wcsa2->ncanonicals) {
		printf ("\n ncanonicals son distintos xxx <%d>, <%d>",wcsa1->ncanonicals, wcsa2->ncanonicals);fflush(stdout);return;
	}		
		//printf ("\n ncanonicals son iguales xxx <%d>, <%d>",wcsa1->ncanonicals, wcsa2->ncanonicals);
		
	for (i=0;i< wcsa1->ncanonicals;i++) {
		if (wcsa1->freqs[i] != wcsa2->freqs[i]) {printf ("freqs son distintos");return;}
		if (wcsa1->lenList[i] != wcsa2->lenList[i]) {printf ("lenList son distintos");return;}
			
	}

	if (wcsa1->occs.bcOffsetSize != wcsa2->occs.bcOffsetSize) {printf ("\n bcOffsetSize size son distintos %d, %d, ",wcsa1->occs.bcOffsetSize,wcsa2->occs.bcOffsetSize);return;}
	if (wcsa1->occs.absOffsetSize != wcsa2->occs.absOffsetSize) {printf ("\n absOffsetSize size son distintos %d, %d, ",wcsa1->occs.absOffsetSize,wcsa2->occs.absOffsetSize);return;}
	if (wcsa1->occs.occsBucketSize != wcsa2->occs.occsBucketSize) {printf ("\n occsBucketSize size son distintos %d, %d, ",wcsa1->occs.occsBucketSize,wcsa2->occs.occsBucketSize);return;}
	if (wcsa1->occs.numBuckets != wcsa2->occs.numBuckets) {printf ("\n numBuckets size son distintos %d, %d, ",wcsa1->occs.numBuckets,wcsa2->occs.numBuckets);return;}
	if (wcsa1->occs.zoneOccsSize != wcsa2->occs.zoneOccsSize) {printf ("\n zoneOccsSize size son distintos %d, %d, ",wcsa1->occs.zoneOccsSize,wcsa2->occs.zoneOccsSize);return;}
	if (wcsa1->occs.sampleK != wcsa2->occs.sampleK) {printf ("\n sampleK size son distintos %d, %d, ",wcsa1->occs.sampleK,wcsa2->occs.sampleK);return;}

	if (wcsa1->occs.bcsSize != wcsa2->occs.bcsSize) {printf ("\n bcs size son distintos %d, %d, ",wcsa1->occs.bcsSize,wcsa2->occs.bcsSize);return;}
	for (i=0; i< wcsa1->occs.bcsSize; i++) {
		if (wcsa1->occs.bcs[i] != wcsa2->occs.bcs[i]) {printf ("bcs son distintos");return;}
	}
	
	if ( (((wcsa1->occs.zoneOccsSize +W-1)/W)-1) != (((wcsa2->occs.zoneOccsSize +W-1)/W)-1))  {printf ("zoneOccsSize son distintos");return;}
	for (i=0; i< (((wcsa1->occs.zoneOccsSize +W-1)/W)-1);i++)
		if (wcsa1->occs.zoneOccs[i] != wcsa2->occs.zoneOccs[i]) {printf ("wcsa->occs.zoneOccs[i] son distintos");return;}
	
	for (i=0; i< wcsa1->ncanonicals; i++) {
		if (wcsa1->occs.occs[i] != wcsa2->occs.occs[i]) {printf ("occs.occs[i] son distintos");return;}
	}	
	
	for (i=0; i< wcsa1->ncanonicals; i++) {
		if (wcsa1->varsCanonicals[i] != wcsa2->varsCanonicals[i]) {printf ("varsCanonicals son distintos");return;}
	}	
	
	if (wcsa1->varsETDCSize !=wcsa2->varsETDCSize) {printf ("varsETDCSizes son distintos");return;}
	for (i=0; i< wcsa1->varsETDCSize; i++) {
		if (wcsa1->varsETDC[i] != wcsa2->varsETDC[i]) {printf ("varsETDC son distintos");return;}
	}	
	
	if (wcsa1->canonicalsZone.size !=wcsa2->canonicalsZone.size) {printf ("canonicalsZone.sizes son distintos");return;}
		printf ("\ncanonicalsZone.sizes son iguales: %d", wcsa1->canonicalsZone.size);		
	for (i=0;i< wcsa1->ncanonicals;i++) {
		if (strcmp( (char *) wcsa1->canonicals[i], (char*) wcsa2->canonicals[i])!=0) {
			printf ("canonicals words son distintos %d <<%s>> <<%s>>",i,wcsa1->canonicals[i],wcsa2->canonicals[i]);return;
			}
	}


	if (wcsa1->wordsZone.size !=wcsa2->wordsZone.size) {printf ("wordsZone.sizes son distintos");return;}		
	for (i=0;i< wcsa1->nwords;i++) {
		if (strcmp( (char *) wcsa1->words[i], (char*) wcsa2->words[i])!=0) {printf ("n words[] son distintos");return;}
	}

	
	printf("\n ****THE INDEXES ARE IDENTICAL \n");
}

