#ifndef II_INCLUDED
#define II_INCLUDED
#include "hashWords.h"

typedef struct {
	uint value;
	uint id;
} tUintSort ;


/** **********************************************************************/
/** Saving/Loading the index to/from disk                               **/
/** **********************************************************************/

		// * loads the compressed text (ETDC) into memory
	int loadTextInMem (byte **text, uint *size, char *basename);
		// * saves the compressed text (ETDC) to disk
	int saveTextInMem (byte *text, uint size, char *basename);
	
		// * Loads the VOCABULARY of ETDC into words array.
	int loadVocabularyETDC (void *index, char* basename);	
		// * Saves the VOCABULARY of ETDC to disk.
	int saveVocabularyETDC (void *index, char* basename);
  
  	// * Creates the array wcsa->indexedWords. (tells if it is valid or not).
	int markIndexedWords (void *index);
  	// * checks if a word in the vocabulary of ETDC is either indexed or not.
 	int isIndexedWord(twcsa *wcsa, uint rank);
 	
 	
		// * Generates a temporary hash table of canonical words.
	void generateTempHashOfCanonicals(Wt_hash *canonicalsHash, void *index);
		// * Calculates the array of sorted canonical words.
	void getSortedCanonicalsArray(Wt_hash canonicalsHash, void *index);
		// * Loads the Ascii of the canonical words from disk
	int loadCanonicalsAscii (void *index, char* basename);	
		// * Loads the Ascii of the canonical words from disk
	int saveCanonicalsAscii (void *index, char* basename);	
	
	// * Assings the codewords of its variants to all canonical words.
	int encodePosOfVariantsOfCanonical (void *index, uint **tmpPosVariants, uint *tmpNumVariants);
  // * Saves the file with the variants of the canonical :: varsETDC y varsCanonicals
	int saveVariantsCanonicals (void *index, char* basename);
  // * Loads the file with the variants of the canonical :: varsETDC y varsCanonicals
	int loadVariantsCanonicals (void *index, char* basename);
	  

		// * Saving freq vector 
	int saveFreqVector(uint *V, uint size, char *basename); 	
		// * Loading  freq vector 
	int loadFreqVector(uint **V, uint *size, char *basename);
	
		// * Saving the lists of occurrences of all the words 
	int saveOccLists (void *index, char *basename);
		// * Loading the lists of occurrences of all the words into "index"
	int loadOccLists (void *index, char *basename);
	
/** **********************************************************************/
/** Some functions used during index construction                        */
/** **********************************************************************/	

  		// * Computes the frequency of each id--word
  		//	* && sets the number of words in the source text: N
	int computeFreqs (void *index, int **mappingToCanonical);
	
		// * Creates the lists of occurrences of all the words
  		// * Block-oriented: Only 1 occurrence per block is marked for each word.
  		// * A word belongs to a block if its associated codeword starts in that block.
  		// * returns also the len of each list.
	//int createListsOfOccurrences(void *index, uint ***occList, uint **lenList);
	int createListsOfOccurrences(void *index, uint ***occList, uint **lenList, int *mapToCanonical);

/** **********************************************************************/
/** Functions needed for searches                                       **/
/** **********************************************************************/	

		// * Uncompresses the list of occurrences of a given id-word
	int uncompressList (void *index, uint id, uint **list, uint *len);
	




#endif
